# Copyright 2017 NTT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import etcd3gw
from oslo_utils import encodeutils


class EtcdClient(object):
    """Define etcd client.

    Only define methods used for networking-spp and make return
    value simple.
    It maybe used 'etcd3' instead of 'etcd3gw' in the future.
    This class also intends to hide incompatibility between
    etcd3 and etcd3gw (when etcd3 will be used).
    """

    def __init__(self, host, port):
        self.client = etcd3gw.client(host, port)

    def safe_decode(self, text):
        if text is None:
            # The watch methods of etcd3gw client can return NoneType for
            # a value when it is not replaced/put but *deleted*. In this
            # case we should return it immediately without decoding.
            return None
        return encodeutils.safe_decode(text)

    def get(self, key):
        data = self.client.get(key)
        if data:
            return self.safe_decode(data[0])

    def get_prefix(self, prefix):
        data = self.client.get_prefix(prefix)
        if data:
            return [(self.safe_decode(meta['key']),
                     self.safe_decode(value)) for value, meta in data]
        return []

    def put(self, key, value):
        self.client.put(key, value)

    def replace(self, key, old_value, new_value):
        return self.client.replace(key, old_value, new_value)

    def delete(self, key):
        return self.client.delete(key)

    def watch_prefix(self, prefix):
        # workaround for etcd3gw bug
        kwargs = {}
        kwargs['range_end'] = \
            etcd3gw.utils._increment_last_byte(prefix)
        orig_iter, cancel = self.client.watch(prefix, **kwargs)

        def iterator():
            for event in orig_iter:
                yield (self.safe_decode(event['kv']['key']),
                       self.safe_decode(event['kv'].get('value')))

        return iterator()

    def watch_once(self, key, timeout=None):
        try:
            event = self.client.watch_once(key, timeout=timeout)
            return (self.safe_decode(event['kv']['key']),
                    self.safe_decode(event['kv'].get('value')))
        except Exception:
            return (None, None)



# Copyright 2017 NTT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import eventlet
eventlet.monkey_patch()

import json
import netaddr
import signal
import sys

from neutron_lib.agent import topics
from neutron_lib import context
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall

from networking_spp.agent import config  # noqa
from networking_spp.agent import spp_api
from networking_spp.common import etcd_client
from networking_spp.common import etcd_config  # noqa
from networking_spp.common import etcd_key
from neutron.agent import rpc as agent_rpc
from neutron.common import config as common_config


LOG = logging.getLogger(__name__)

AGENT_TYPE_SPP = 'SPP neutron agent'
SPP_AGENT_BINARY = 'neutron-spp-agent'


class Vhostuser(object):

    def __init__(self, name, vf):
        self.name = name
        self.vf = vf
        self.phys_net = vf.phys_net
        self.mac_address = "unuse"

        # port name and component name handled for del_vlantag
        self.del_vlan_port = None
        self.del_vlan_comp = None

        # port name and component name handled for add_vlantag
        self.add_vlan_port = None
        self.add_vlan_comp = None

        # setting for taas
        # component name handled when vhost is service port
        self.dst_comp = None

        # port(ring) name and component name handled when vhost is source port
        self.in_ring = None
        self.in_comp = None
        self.out_ring = None
        self.out_comp = None


class SppVf(spp_api.SppVfApi):

    def __init__(self, sec_id, phys_net, api_ip_addr, api_port):
        super(SppVf, self).__init__(sec_id, api_ip_addr, api_port)

        self.phys_net = phys_net
        self.vhostusers = {}

    def init_components(self, components):
        self.get_status()
        self.build_vhosts(components)
        self.build_components(components)
        self.init_vhost_mac_address()

    def build_components(self, components):
        exist_comps = {}
        for comp in self.info["components"]:
            if comp["type"] != "unuse":
                exist_comps[comp["name"]] = comp
                # backward compatibility for yaml definition.
                if comp["type"] == "classifier_mac":
                    LOG.warning("type 'classifier_mac' is obsoleted,"
                                " use 'classifier'.")
                    comp["type"] = "classifier"

        for comp in components:
            comp_name = comp["name"]
            if comp_name not in exist_comps:
                # if component does not exist, make component and add its
                # ports according to the configuration.
                self.make_component(comp_name, comp["core"], comp["type"])
                for port in comp["tx_port"]:
                    self.port_add(port, "tx", comp_name)
                for port in comp["rx_port"]:
                    self.port_add(port, "rx", comp_name)
                continue

            # if component exists, add its ports if not added yet.
            # the complication is that wiring may be different from
            # the configuration by taas operation.
            exist_rx_port = []
            exist_tx_port = []
            for port in exist_comps[comp_name]["rx_port"]:
                exist_rx_port.append(port["port"])
            for port in exist_comps[comp_name]["tx_port"]:
                exist_tx_port.append(port["port"])

            if comp["type"] == "classifier":
                # classifier is not affected by taas.
                # add ports according to the configuration if not added.
                for port in comp["tx_port"]:
                    if port not in exist_tx_port:
                        self.port_add(port, "tx", comp_name)
                if not exist_rx_port:
                    port = comp["rx_port"][0]
                    self.port_add(port, "rx", comp_name)
            elif comp["type"] == "forward":
                # tx_port may differ from the configuration by taas.
                # if not added, add a port according to the configuration.
                # note that if it was under taas operation, it will be
                # recovered later.
                if not exist_tx_port:
                    port = comp["tx_port"][0]
                    self.port_add(port, "tx", comp_name)
                # rx_port is not affected by taas.
                if not exist_rx_port:
                    port = comp["rx_port"][0]
                    self.port_add(port, "rx", comp_name)
            else:  # "type" == "merge"
                # merge type is used the following two cases.
                # * forwarder of vhost: tx_port is vhost
                # * merger to physical nic: tx_port is phys
                port = comp["tx_port"][0]
                # tx_port is not affected by taas both cases.
                if not exist_tx_port:
                    self.port_add(port, "tx", comp_name)
                # rx_port
                if port.startswith("vhost"):
                    # rx_port may differ from the configuration by taas.
                    # if not added, add a port according to the configuration.
                    # note that if it was under taas operation, it will be
                    # recovered later.
                    if not exist_rx_port:
                        port = comp["rx_port"][0]
                        self.port_add(port, "rx", comp_name)
                else:  # phys
                    # this case is not affected by taas.
                    # add ports according to the configuration if not added.
                    for port in comp["rx_port"]:
                        if port not in exist_rx_port:
                            self.port_add(port, "rx", comp_name)

        # to output info after build to debug LOG.
        self.get_status()

    def build_vhosts(self, components):
        tx_port_to_comp = {}
        rx_port_to_comp = {}
        for comp in components:
            for port in comp["tx_port"]:
                tx_port_to_comp[port] = comp
            for port in comp["rx_port"]:
                rx_port_to_comp[port] = comp

        for port in tx_port_to_comp.keys():
            if_type, if_num = port.split(":")
            if if_type == "vhost":
                if_num = int(if_num)
                self.vhostusers[if_num] = Vhostuser(port, self)

        for vhost in self.vhostusers.values():
            # tx side component of vhost
            comp = tx_port_to_comp[vhost.name]
            if comp["type"] == "forward":
                # change type forward to merge for mirror(taas) support.
                # there is no problem to use merge even if taas is not used.
                comp["type"] = "merge"
            if comp["type"] == "merge":
                #                          del_vlan target
                # vhost -- tx[merge]rx -- ring -- tx[classifier]rx -- phys
                #
                vhost.del_vlan_port = comp["rx_port"][0]
                vhost.del_vlan_comp = (tx_port_to_comp[vhost.del_vlan_port]
                                       ["name"])
                # setting for taas
                vhost.dst_comp = comp["name"]
                vhost.in_comp = comp["name"]
                vhost.in_ring = vhost.del_vlan_port
            else:  # "type" == "classifier"
                #   del_vlan target
                # vhost -- tx[classifier]rx -- phys
                #
                vhost.del_vlan_port = vhost.name
                vhost.del_vlan_comp = comp["name"]
                # note: taas is not supported this configuration.

            # rx side component of vhost
            comp = rx_port_to_comp[vhost.name]
            if comp["type"] == "forward":
                #                            add_vlan target
                # vhost -- rx[forward]tx -- ring -- rx[merge]tx -- phys
                #
                vhost.add_vlan_port = comp["tx_port"][0]
                vhost.add_vlan_comp = (rx_port_to_comp[vhost.add_vlan_port]
                                       ["name"])
                # setting for taas
                vhost.out_comp = comp["name"]
                vhost.out_ring = vhost.add_vlan_port
            else:  # "type" == "merge"
                #  add_vlan target
                # vhost -- rx[merge]tx -- phys
                #
                vhost.add_vlan_port = vhost.name
                vhost.add_vlan_comp = comp["name"]
                # note: taas is not supported this configuration.

    def init_vhost_mac_address(self):
        table = self.info.get("classifier_table", [])
        for vhost in self.vhostusers.values():
            for entry in table:
                if (entry["type"] in ["mac", "vlan"] and
                        entry["port"] == vhost.del_vlan_port):
                    mac_string = entry["value"]
                    if entry["type"] == "vlan":
                        _vlan_id, mac_string = mac_string.split('/')
                    # match format of neutron standard
                    mac = str(netaddr.EUI(mac_string,
                                          dialect=netaddr.mac_unix_expanded))
                    vhost.mac_address = mac


class Mirror(object):

    def __init__(self, comp, ports, proc):
        self.comp = comp
        self.proc = proc
        self.ring_a = ports[0]  # connect to service side
        self.ring_b = ports[1]  # connect to source side


class SppMirror(spp_api.SppMirrorApi):

    def __init__(self, sec_id, api_ip_addr, api_port):
        super(SppMirror, self).__init__(sec_id, api_ip_addr, api_port)

        self.mirrors = []

    def _num_to_name(self, num):
        return "mirror_%d" % num

    def init_components(self, components):
        self.get_status()

        exist_comps = {}
        for comp in self.info["components"]:
            if comp["type"] != "unuse":
                exist_comps[comp["name"]] = comp

        for i in range(len(components)):
            comp_name = self._num_to_name(i)
            if comp_name not in exist_comps:
                core_id = components[i]["core"]
                self.make_component(comp_name, core_id)
            ports = components[i]["ports"]
            self.mirrors.append(Mirror(comp_name, ports, self))


class SppAgent(object):

    def __init__(self, conf):

        self.conf = conf
        self.host = self.conf.host
        self.api_ip_addr = self.conf.spp.api_ip_addr
        self.api_port = self.conf.spp.api_port
        self.etcd = etcd_client.EtcdClient(self.conf.spp.etcd_host,
                                           self.conf.spp.etcd_port)
        self.spp_configuration = self.get_spp_configuration()

        self.vhostusers = {}
        sec_id = 1
        for mapping in self.spp_configuration['vf']:
            phys_net = mapping['physical_network']
            components = mapping['components']
            vf = SppVf(sec_id, phys_net, self.api_ip_addr, self.api_port)
            vf.init_components(components)
            self.vhostusers.update(vf.vhostusers)
            sec_id += 1

        self.mirror = None
        mirror_components = self.spp_configuration.get('mirror')
        if mirror_components:
            self.mirror = SppMirror(sec_id, self.api_ip_addr, self.api_port)
            self.mirror.init_components(mirror_components)

        self.plug_sem = eventlet.semaphore.Semaphore()
        self.shutdown_sem = eventlet.semaphore.Semaphore(value=0)
        self.port_plug_watch_failed = False
        eventlet.spawn_n(self.port_plug_watch)
        # to start port_plug_watch first
        eventlet.sleep(0)
        if self.mirror:
            eventlet.spawn_n(self.tap_plug_watch)
            eventlet.sleep(0)
        self.recover()
        self.start_report()

    def get_spp_configuration(self):
        value = self.etcd.get(etcd_key.configuration_key(self.host))
        mappings = json.loads(value)
        LOG.info("SPP configuration: %s", mappings)
        return mappings

    def set_classifier_table(self, vhost_id, mac_address, vlan_id):
        vhost = self.vhostusers[vhost_id]
        if vhost.mac_address == mac_address:
            LOG.debug("classifier table already set: %d: %s",
                      vhost_id, mac_address)
            return

        vf = vhost.vf
        if vlan_id is None:
            vf.set_classifier_table(mac_address, vhost.del_vlan_port)
        else:
            vf.port_del(vhost.add_vlan_port, "rx", vhost.add_vlan_comp)
            vf.port_add(vhost.add_vlan_port, "rx", vhost.add_vlan_comp,
                        "add_vlantag", vlan_id)
            vf.port_del(vhost.del_vlan_port, "tx", vhost.del_vlan_comp)
            vf.port_add(vhost.del_vlan_port, "tx", vhost.del_vlan_comp,
                        "del_vlantag")
            vf.set_classifier_table_with_vlan(mac_address,
                                              vhost.del_vlan_port,
                                              vlan_id)
        vhost.mac_address = mac_address

    def clear_classifier_table(self, vhost_id, mac_address, vlan_id):
        vhost = self.vhostusers[vhost_id]
        if vhost.mac_address == "unuse":
            LOG.debug("classifier table already clear: %d", vhost_id)
            return

        vf = vhost.vf
        if vlan_id is None:
            vf.clear_classifier_table(mac_address, vhost.del_vlan_port)
        else:
            vf.port_del(vhost.add_vlan_port, "rx", vhost.add_vlan_comp)
            vf.port_add(vhost.add_vlan_port, "rx", vhost.add_vlan_comp)
            vf.port_del(vhost.del_vlan_port, "tx", vhost.del_vlan_comp)
            vf.port_add(vhost.del_vlan_port, "tx", vhost.del_vlan_comp)
            vf.clear_classifier_table_with_vlan(mac_address,
                                                vhost.del_vlan_port,
                                                vlan_id)
        vhost.mac_address = "unuse"

    def _plug_port(self, port_id, vhost_id, mac_address, vlan_id):
        LOG.info("plug port %s: mac: %s, vhost: %d, vlan_id: %s", port_id,
                 mac_address, vhost_id, vlan_id)

        self.set_classifier_table(vhost_id, mac_address, vlan_id)

        key = etcd_key.port_status_key(self.host, port_id)
        self.etcd.put(key, "up")

    def _unplug_port(self, port_id, vhost_id, mac_address, vlan_id):
        LOG.info("unplug port %s: mac: %s, vhost: %d, vlan_id: %s", port_id,
                 mac_address, vhost_id, vlan_id)

        if self.mirror:
            self._unplug_tap_port(port_id)

        self.clear_classifier_table(vhost_id, mac_address, vlan_id)

        phys_net = self.vhostusers[vhost_id].phys_net
        key = etcd_key.vhost_key(self.host, phys_net, vhost_id)
        self.etcd.replace(key, port_id, 'None')
        delete_keys = [etcd_key.bind_port_key(self.host, port_id),
                       etcd_key.action_key(self.host, port_id),
                       etcd_key.port_status_key(self.host, port_id)]
        for key in delete_keys:
            self.etcd.delete(key)

    def _do_plug_unplug(self, port_id):
        with self.plug_sem:
            value = self.etcd.get(etcd_key.bind_port_key(self.host, port_id))
            if value is None:
                # this can happen under port_plug_watch and recover race
                # condition.
                LOG.debug("already deleted %s", port_id)
                return
            data = json.loads(value)
            vhost_id = data['vhost_id']
            mac_address = data['mac_address']
            vlan_id = data.get('vlan_id')
            # get action again to get the newest value
            op = self.etcd.get(etcd_key.action_key(self.host, port_id))
            if op == "plug":
                self._plug_port(port_id, vhost_id, mac_address, vlan_id)
            else:
                self._unplug_port(port_id, vhost_id, mac_address, vlan_id)

    def port_plug_watch(self):
        LOG.info("SPP port_plug_watch stated")
        prefix = etcd_key.action_host_prefix(self.host)
        try:
            for key, value in self.etcd.watch_prefix(prefix):
                if value in ["plug", "unplug"]:
                    LOG.debug("SPP port_plug_watch: %s %s", key, value)
                    port_id = key[len(prefix):]
                    self._do_plug_unplug(port_id)
        except Exception as e:
            # basically operation should not be failed.
            # it may be critical error. so shutdown agent.
            LOG.error("Failed to plug/unplug: %s", e)
            self.port_plug_watch_failed = True
            self.shutdown_sem.release()

    def _port_id_to_vhost(self, port_id):
        key = etcd_key.bind_port_key(self.host, port_id)
        val = self.etcd.get(key)
        if val:
            data = json.loads(val)
            return self.vhostusers[data['vhost_id']]

    def _ring_add(self, proc, ring, direction, comp):
        if not proc.port_exist(ring, direction, comp):
            proc.port_add(ring, direction, comp)

    def _ring_del(self, proc, ring, direction, comp):
        if proc.port_exist(ring, direction, comp):
            proc.port_del(ring, direction, comp)

    def _attach_ring(self, ring, rx_proc, rx_comp, tx_proc, tx_comp):
        self._ring_add(rx_proc, ring, "rx", rx_comp)
        self._ring_add(tx_proc, ring, "tx", tx_comp)

    def _detach_ring(self, ring, rx_proc, rx_comp, tx_proc, tx_comp):
        self._ring_del(tx_proc, ring, "tx", tx_comp)
        self._ring_del(rx_proc, ring, "rx", rx_comp)

    def _change_connection(self, ring, direction,
                           del_proc, del_comp, add_proc, add_comp):
        self._ring_del(del_proc, ring, direction, del_comp)
        self._ring_add(add_proc, ring, direction, add_comp)

    # tap-in construction
    #
    #                       rx ---(original)
    #                    +--------+
    # dst_vhost rx --- tx|dst_comp|
    #                    +--------+ (1)attach
    #                       rx --- ring_a --- tx
    #                                       +------+
    #                                       |mirror|rx --+
    #                        (2)attach      +------+     |
    #                       rx --- ring_b --- tx         |
    #                    +--------+                      |
    # src_vhost rx --- tx|in_comp |rx ...................+-- in_ring --tx[..]
    #                    +--------+    (3)change connection
    #
    def _construct_tap_in(self, mirror, dst_vhost, src_vhost):
        self._attach_ring(mirror.ring_a,
                          dst_vhost.vf, dst_vhost.dst_comp,
                          mirror.proc, mirror.comp)
        self._attach_ring(mirror.ring_b,
                          src_vhost.vf, src_vhost.in_comp,
                          mirror.proc, mirror.comp)
        self._change_connection(src_vhost.in_ring, "rx",
                                src_vhost.vf, src_vhost.in_comp,
                                mirror.proc, mirror.comp)

    # tap-in destruction: reverse operation of construction
    def _destruct_tap_in(self, mirror, dst_vhost, src_vhost):
        self._change_connection(src_vhost.in_ring, "rx",
                                mirror.proc, mirror.comp,
                                src_vhost.vf, src_vhost.in_comp)
        self._detach_ring(mirror.ring_b,
                          src_vhost.vf, src_vhost.in_comp,
                          mirror.proc, mirror.comp)
        self._detach_ring(mirror.ring_a,
                          dst_vhost.vf, dst_vhost.dst_comp,
                          mirror.proc, mirror.comp)

    # tap-out construction
    #
    #                       rx ---(original)
    #                    +--------+
    # dst_vhost rx --- tx|dst_comp|
    #                    +--------+ (1)attach
    #                       rx --- ring_a --- tx
    #                                       +------+
    #                               +---- rx|mirror|tx --+
    #                   (3)attach ring_b    +------+     |
    #                               |                    |
    #                    +--------+ +                    |
    # src_vhost tx --- rx|out_comp|tx ...................+-- out_ring --rx[..]
    #                    +--------+    (2)change connection
    #
    def _construct_tap_out(self, mirror, dst_vhost, src_vhost):
        self._attach_ring(mirror.ring_a,
                          dst_vhost.vf, dst_vhost.dst_comp,
                          mirror.proc, mirror.comp)
        self._change_connection(src_vhost.out_ring, "tx",
                                src_vhost.vf, src_vhost.out_comp,
                                mirror.proc, mirror.comp)
        self._attach_ring(mirror.ring_b,
                          mirror.proc, mirror.comp,
                          src_vhost.vf, src_vhost.out_comp)

    # tap-out destruction: reverse operation of construction
    def _destruct_tap_out(self, mirror, dst_vhost, src_vhost):
        self._detach_ring(mirror.ring_b,
                          mirror.proc, mirror.comp,
                          src_vhost.vf, src_vhost.out_comp)
        self._change_connection(src_vhost.out_ring, "tx",
                                mirror.proc, mirror.comp,
                                src_vhost.vf, src_vhost.out_comp)
        self._detach_ring(mirror.ring_a,
                          dst_vhost.vf, dst_vhost.dst_comp,
                          mirror.proc, mirror.comp)

    def _plug_tap(self, tap_flow_id, mirror_in, mirror_out, dst_vhost,
                  src_vhost):
        if mirror_in is not None:
            self._construct_tap_in(self.mirror.mirrors[mirror_in],
                                   dst_vhost, src_vhost)

        if mirror_out is not None:
            self._construct_tap_out(self.mirror.mirrors[mirror_out],
                                    dst_vhost, src_vhost)

        key = etcd_key.tap_status_key(self.host, tap_flow_id)
        self.etcd.put(key, "up")

    def _unplug_tap(self, tap_flow_id, mirror_in, mirror_out, dst_vhost,
                    src_vhost):
        if mirror_in is not None:
            self._destruct_tap_in(self.mirror.mirrors[mirror_in],
                                  dst_vhost, src_vhost)
            key = etcd_key.mirror_key(self.host, mirror_in)
            self.etcd.put(key, 'None')

        if mirror_out is not None:
            self._destruct_tap_out(self.mirror.mirrors[mirror_out],
                                   dst_vhost, src_vhost)
            key = etcd_key.mirror_key(self.host, mirror_out)
            self.etcd.put(key, 'None')

        delete_keys = [etcd_key.tap_info_key(self.host, tap_flow_id),
                       etcd_key.tap_action_key(self.host, tap_flow_id),
                       etcd_key.tap_status_key(self.host, tap_flow_id)]
        for key in delete_keys:
            self.etcd.delete(key)

    def _do_tap_plug_unplug(self, tap_flow_id):
        with self.plug_sem:
            key = etcd_key.tap_info_key(self.host, tap_flow_id)
            value = self.etcd.get(key)
            if value is None:
                # this can happen under tap_plug_watch and recover race
                # condition.
                LOG.debug("already deleted %s", tap_flow_id)
                return
            tap_info = json.loads(value)
            mirror_in = tap_info['mirror_in']
            mirror_out = tap_info['mirror_out']
            dst_vhost = self._port_id_to_vhost(tap_info['service_port'])
            src_vhost = self._port_id_to_vhost(tap_info['source_port'])
            if dst_vhost is None or src_vhost is None:
                return
            self.mirror.get_status()
            dst_vhost.vf.get_status()
            if src_vhost.vf != dst_vhost.vf:
                src_vhost.vf.get_status()
            # get action again to get the newest value
            op = self.etcd.get(etcd_key.tap_action_key(self.host, tap_flow_id))
            if op == "plug":
                self._plug_tap(tap_flow_id, mirror_in, mirror_out,
                               dst_vhost, src_vhost)
            else:
                self._unplug_tap(tap_flow_id, mirror_in, mirror_out,
                                 dst_vhost, src_vhost)

    def _unplug_tap_port(self, port_id):
        prefix = etcd_key.tap_info_host_prefix(self.host)
        for key, value in self.etcd.get_prefix(prefix):
            tap_flow_id = key[len(prefix):]
            tap_info = json.loads(value)
            if (port_id != tap_info['service_port'] and
                    port_id != tap_info['source_port']):
                continue
            mirror_in = tap_info['mirror_in']
            mirror_out = tap_info['mirror_out']
            dst_vhost = self._port_id_to_vhost(tap_info['service_port'])
            src_vhost = self._port_id_to_vhost(tap_info['source_port'])
            if dst_vhost is None or src_vhost is None:
                continue
            self.mirror.get_status()
            dst_vhost.vf.get_status()
            if src_vhost.vf != dst_vhost.vf:
                src_vhost.vf.get_status()
            self._unplug_tap(tap_flow_id, mirror_in, mirror_out,
                             dst_vhost, src_vhost)

    def tap_plug_watch(self):
        LOG.info("SPP tap_plug_watch stated")
        prefix = etcd_key.tap_action_host_prefix(self.host)
        try:
            for key, value in self.etcd.watch_prefix(prefix):
                if value in ["plug", "unplug"]:
                    LOG.debug("SPP tap_plug_watch: %s %s", key, value)
                    tap_flow_id = key[len(prefix):]
                    self._do_tap_plug_unplug(tap_flow_id)
        except Exception as e:
            # basically operation should not be failed.
            # it may be critical error. so shutdown agent.
            LOG.error("Failed to tap plug/unplug: %s", e)
            self.port_plug_watch_failed = True
            self.shutdown_sem.release()

    def recover(self):
        LOG.debug("recover start")
        prefix = etcd_key.action_host_prefix(self.host)
        for key, value in self.etcd.get_prefix(prefix):
            LOG.debug("  %s: %s", key, value)
            if value in ["plug", "unplug"]:
                port_id = key[len(prefix):]
                self._do_plug_unplug(port_id)

        prefix = etcd_key.tap_action_host_prefix(self.host)
        for key, value in self.etcd.get_prefix(prefix):
            LOG.debug("  %s: %s", key, value)
            if value in ["plug", "unplug"]:
                tap_flow_id = key[len(prefix):]
                self._do_tap_plug_unplug(tap_flow_id)

    def _handle_signal(self, signum, frame):
        LOG.info("signal recieved.")
        self.shutdown_sem.release()

    def wait_shutdown(self):
        LOG.info("Agent initialized successfully, now running... ")

        signal.signal(signal.SIGTERM, self._handle_signal)
        signal.signal(signal.SIGINT, self._handle_signal)

        # sleep until signal recieved
        self.shutdown_sem.acquire()

        LOG.info("Agent shutdown done")
        if self.port_plug_watch_failed:
            return 1
        else:
            return 0

    def start_report(self):
        self.state_rpc = agent_rpc.PluginReportStateAPI(topics.REPORTS)
        self.context = context.get_admin_context_without_session()
        self.agent_state = {
            'binary': SPP_AGENT_BINARY,
            'host': self.host,
            'topic': 'N/A',
            'agent_type': AGENT_TYPE_SPP,
            'start_flag': True}

        report_interval = self.conf.AGENT.report_interval
        if report_interval:
            heartbeat = loopingcall.FixedIntervalLoopingCall(
                self._report_state)
            heartbeat.start(interval=report_interval)

    def _report_state(self):
        try:
            self.state_rpc.report_state(self.context,
                                        self.agent_state)
            self.agent_state.pop('start_flag', None)
        except Exception:
            LOG.exception("Failed reporting state!")


def main():
    common_config.init(sys.argv[1:])
    common_config.setup_logging()
    try:
        agent = SppAgent(cfg.CONF)
    except Exception:
        LOG.exception("Agent Initialization Failed")
        return 1
    return agent.wait_shutdown()


if __name__ == "__main__":
    sys.exit(main())



# Copyright (c) 2017 NTT
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from tempest.test_discover import plugins


class NetworkingSppTempestPlugin(plugins.TempestPlugin):
    def get_opt_lists(self):
        return []

    def load_tests(self):
        # .../networking-spp/networking_spp/tests/tempest/plugin.py
        #     top_dir        d2             d1    test_dir __file__
        test_dir = os.path.dirname(os.path.abspath(__file__))
        d1 = os.path.dirname(test_dir)
        d2 = os.path.dirname(d1)
        top_dir = os.path.dirname(d2)
        return (test_dir, top_dir)

    def register_opts(self, conf):
        return



# Copyright 2017 NTT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


SPP_ROOT = "/spp/openstack/"

# configuration of each agent
# /spp/openstack/configuration/<host>
SPP_PREFIX_CONF = SPP_ROOT + "configuration/"
SPP_KEY_CONF = SPP_PREFIX_CONF + "{host}"

# manage vhost usage
# /spp/openstack/vhost/<host>/<phys>/<vhost_id>
SPP_PREFIX_VHOST = SPP_ROOT + "vhost/"
SPP_PREFIX_VHOST_HOST = SPP_PREFIX_VHOST + "{host}/"
SPP_PREFIX_VHOST_HOST_PHYS = SPP_PREFIX_VHOST_HOST + "{phys}/"
SPP_KEY_VHOST = SPP_PREFIX_VHOST_HOST_PHYS + "{vhost_id}"

# to indicate plug done
# /spp/openstack/port_status/<host>/<port id>
SPP_PREFIX_PORT_STATUS = SPP_ROOT + "port_status/"
SPP_PREFIX_PORT_STATUS_HOST = SPP_PREFIX_PORT_STATUS + "{host}/"
SPP_KEY_PORT_STATUS = SPP_PREFIX_PORT_STATUS_HOST + "{port_id}"

# port info for plug
# /spp/openstack/bind_port/<host>/<port id>
SPP_PREFIX_BIND_PORT = SPP_ROOT + "bind_port/"
SPP_PREFIX_PORT_BIND_PORT_HOST = SPP_PREFIX_BIND_PORT + "{host}/"
SPP_KEY_BIND_PORT = SPP_PREFIX_PORT_BIND_PORT_HOST + "{port_id}"

# plug/unplug port
# /spp/openstack/action/<host>/<port id>
SPP_PREFIX_ACTION = SPP_ROOT + "action/"
SPP_PREFIX_ACTION_HOST = SPP_PREFIX_ACTION + "{host}/"
SPP_KEY_ACTION = SPP_PREFIX_ACTION_HOST + "{port_id}"

# manage mirror usage
# /spp/openstack/mirror/<host>/<mirror_id>
SPP_PREFIX_MIRROR = SPP_ROOT + "mirror/"
SPP_PREFIX_MIRROR_HOST = SPP_PREFIX_MIRROR + "{host}/"
SPP_KEY_MIRROR = SPP_PREFIX_MIRROR_HOST + "{mirror_id}"

# to indicate tap plug done
# /spp/openstack/tap_status/<host>/<tap_flow_id>
SPP_PREFIX_TAP_STATUS = SPP_ROOT + "tap_status/"
SPP_PREFIX_TAP_STATUS_HOST = SPP_PREFIX_TAP_STATUS + "{host}/"
SPP_KEY_TAP_STATUS = SPP_PREFIX_TAP_STATUS_HOST + "{tap_flow_id}"

# info for tap plug
# /spp/openstack/tap_info/<host>/<tap_flow_id>
SPP_PREFIX_TAP_INFO = SPP_ROOT + "tap_info/"
SPP_PREFIX_TAP_INFO_HOST = SPP_PREFIX_TAP_INFO + "{host}/"
SPP_KEY_TAP_INFO = SPP_PREFIX_TAP_INFO_HOST + "{tap_flow_id}"

# plug/unplug tap
# /spp/openstack/tap_action/<host>/<tap_flow_id>
SPP_PREFIX_TAP_ACTION = SPP_ROOT + "tap_action/"
SPP_PREFIX_TAP_ACTION_HOST = SPP_PREFIX_TAP_ACTION + "{host}/"
SPP_KEY_TAP_ACTION = SPP_PREFIX_TAP_ACTION_HOST + "{tap_flow_id}"


def configuration_key(host):
    return SPP_KEY_CONF.format(**locals())


def vhost_phys_prefix(host, phys):
    return SPP_PREFIX_VHOST_HOST_PHYS.format(**locals())


def vhost_key(host, phys, vhost_id):
    return SPP_KEY_VHOST.format(**locals())


def port_status_prefix():
    return SPP_PREFIX_PORT_STATUS


def port_status_key(host, port_id):
    return SPP_KEY_PORT_STATUS.format(**locals())


def bind_port_key(host, port_id):
    return SPP_KEY_BIND_PORT.format(**locals())


def action_host_prefix(host):
    return SPP_PREFIX_ACTION_HOST.format(**locals())


def action_key(host, port_id):
    return SPP_KEY_ACTION.format(**locals())


def mirror_prefix(host):
    return SPP_PREFIX_MIRROR_HOST.format(**locals())


def mirror_key(host, mirror_id):
    return SPP_KEY_MIRROR.format(**locals())


def tap_status_key(host, tap_flow_id):
    return SPP_KEY_TAP_STATUS.format(**locals())


def tap_info_host_prefix(host):
    return SPP_PREFIX_TAP_INFO_HOST.format(**locals())


def tap_info_key(host, tap_flow_id):
    return SPP_KEY_TAP_INFO.format(**locals())


def tap_action_host_prefix(host):
    return SPP_PREFIX_TAP_ACTION_HOST.format(**locals())


def tap_action_key(host, tap_flow_id):
    return SPP_KEY_TAP_ACTION.format(**locals())



# Copyright (c) 2017 NTT
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from networking_spp.common import etcd_client
from networking_spp.common import etcd_config  # noqa
from networking_spp.common import etcd_key
from neutron_lib.api.definitions import portbindings
from neutron_lib import constants
from neutron_lib import context
from neutron_lib.plugins.ml2 import api
from oslo_config import cfg
from oslo_log import log


LOG = log.getLogger(__name__)

# TODO(oda): define in common place
AGENT_TYPE_SPP = 'SPP neutron agent'
WAIT_PLUG_TIMEOUT = 20


class SppMechanismDriver(api.MechanismDriver):

    def initialize(self):
        self.context = context.get_admin_context()
        self.etcd = etcd_client.EtcdClient(cfg.CONF.spp.etcd_host,
                                           cfg.CONF.spp.etcd_port)

    def _try_to_bind(self, context, segment):
        port_id = context.current['id']
        if self.etcd.get(etcd_key.bind_port_key(context.host, port_id)):
            # check already tried to bind at first.
            return
        phys = segment[api.PHYSICAL_NETWORK]
        prefix = etcd_key.vhost_phys_prefix(context.host, phys)
        vhost_id = None
        for key, value in self.etcd.get_prefix(prefix):
            if value == 'None':
                if self.etcd.replace(key, 'None', port_id):
                    vhost_id = key[len(prefix):]
                    LOG.debug("vhost %s assigned for port %s", vhost_id,
                              port_id)
                    break
        if not vhost_id:
            LOG.warn("no vhost available for port %s", port_id)
            return

        self._add_bind_port(context, vhost_id, segment[api.SEGMENTATION_ID])
        if not self._wait_plug_port(context.host, port_id):
            return

        vif_type = 'vhostuser'
        sock_path = "/tmp/sock%s" % vhost_id
        mode = portbindings.VHOST_USER_MODE_SERVER
        vif_details = {portbindings.CAP_PORT_FILTER: False,
                       portbindings.VHOST_USER_MODE: mode,
                       portbindings.VHOST_USER_SOCKET: sock_path}
        context.set_binding(segment[api.ID], vif_type, vif_details,
                            status=constants.PORT_STATUS_ACTIVE)

    def _spp_agent_alive(self, context):
        agents = context.host_agents(AGENT_TYPE_SPP)
        if agents:
            return agents[0]['alive']
        return False

    def bind_port(self, context):
        agent_conf = self.etcd.get(etcd_key.configuration_key(context.host))
        if not agent_conf:
            # spp is not configured on the host
            return
        agent_conf = json.loads(agent_conf)
        phys_nets = [conf['physical_network'] for conf in agent_conf['vf']]
        if not self._spp_agent_alive(context):
            LOG.warn("spp_agent of %s down", context.host)
            return

        for segment in context.segments_to_bind:
            if (segment[api.NETWORK_TYPE] in [constants.TYPE_FLAT,
                                              constants.TYPE_VLAN]
                    and segment[api.PHYSICAL_NETWORK] in phys_nets):
                self._try_to_bind(context, segment)
                return

    def _add_bind_port(self, context, vhost_id, vlan_id):
        port = context.current
        value = {'mac_address': port['mac_address'], 'vhost_id': int(vhost_id)}
        if vlan_id is not None:
            value['vlan_id'] = vlan_id
        value = json.dumps(value)
        key = etcd_key.bind_port_key(context.host, port['id'])
        self.etcd.put(key, value)
        key = etcd_key.action_key(context.host, port['id'])
        self.etcd.put(key, "plug")

    def _wait_plug_port(self, host, port_id):
        key = etcd_key.port_status_key(host, port_id)
        k, v = self.etcd.watch_once(key, timeout=WAIT_PLUG_TIMEOUT)
        if k and v == "up":
            LOG.debug("port %s plug done.", port_id)
            return True

    def _unplug_port(self, host, port_id):
        key = etcd_key.action_key(host, port_id)
        value = self.etcd.get(key)
        if value:
            self.etcd.put(key, "unplug")

    def update_port_postcommit(self, context):
        if (context.original_host and not context.host
                and context.original_vif_type in ('vhostuser',
                                                  'binding_failed')):
            self._unplug_port(context.original_host, context.current['id'])

    def delete_port_postcommit(self, context):
        if (context.host
                and context.vif_type in ('vhostuser', 'binding_failed')):
            self._unplug_port(context.host, context.current['id'])



# Copyright 2017 NTT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import etcd3
import json
import os
import sys
import yaml

SPP_ROOT = "/spp/openstack/"


def config_path(host):
    return SPP_ROOT + "configuration/" + host


def vhost_path(host, phys_net, vhost):
    return SPP_ROOT + "vhost/%s/%s/%d" % (host, phys_net, vhost)


def mirror_path(host, mirror_id):
    return SPP_ROOT + "mirror/%s/%d" % (host, mirror_id)


def get_cores(core_mask):
    mask = int(core_mask, base=16)
    cores = []
    for i in range(32):
        if mask & (1 << i):
            cores.append(i)
    cores.pop(0)
    return cores


# port name conventions
def _nic_port(sec_id):
    return "phy:%d" % (sec_id - 1)


def _vhost_port(vhost_id):
    return "vhost:%d" % vhost_id


def _rx_ring_port(vhost_id):
    return "ring:%d" % (vhost_id * 2)


def _tx_ring_port(vhost_id):
    return "ring:%d" % (vhost_id * 2 + 1)


def conf_components(sec_id, start_vhost_id, num_vhost, cores):
    components = []
    tx_port = []
    rx_port = []
    for vhost_id in range(start_vhost_id, start_vhost_id + num_vhost):
        components.append({"core": cores.pop(0),
                           "type": "forward",
                           "name": "forward_%d_tx" % vhost_id,
                           "rx_port": [_rx_ring_port(vhost_id)],
                           "tx_port": [_vhost_port(vhost_id)]})
        components.append({"core": cores.pop(0),
                           "type": "forward",
                           "name": "forward_%d_rx" % vhost_id,
                           "rx_port": [_vhost_port(vhost_id)],
                           "tx_port": [_tx_ring_port(vhost_id)]})

        tx_port.append(_rx_ring_port(vhost_id))
        rx_port.append(_tx_ring_port(vhost_id))

    components.append({"core": cores.pop(0),
                       "type": "classifier",
                       "name": "classifier",
                       "rx_port": [_nic_port(sec_id)],
                       "tx_port": tx_port})
    components.append({"core": cores.pop(0),
                       "type": "merge",
                       "name": "merger",
                       "rx_port": rx_port,
                       "tx_port": [_nic_port(sec_id)]})
    return components


def mirror_components(num_vhost, num_mirror, cores):
    ring_id = num_vhost * 2
    components = []
    for i in range(num_mirror):
        components.append({"core": cores.pop(0),
                           "ports": ["ring:%d" % ring_id,
                                     "ring:%d" % (ring_id + 1)]})
        ring_id += 2
    return components


def main():
    dpdk_port_mappings = os.environ.get("DPDK_PORT_MAPPINGS")
    host = os.environ.get("SPP_HOST")
    etcd_host = os.environ.get("ETCD_HOST")
    etcd_port = os.environ.get("ETCD_PORT")
    if (dpdk_port_mappings is None or host is None or
            etcd_host is None or etcd_port is None):
        print("DPDK_PORT_MAPPINGS, SPP_HOST, ETCD_HOST and ETCD_PORT"
              " must be defined.")
        return 1
    component_conf = os.environ.get("SPP_COMPONENT_CONF")
    spp_mirror = os.environ.get("SPP_MIRROR")

    def_vfs = []
    def_mirror = {}
    if component_conf:
        with open(component_conf, 'r') as f:
            def_confs = yaml.load(f)
            def_vfs = def_confs['vf']
            if spp_mirror:
                def_mirror = def_confs['mirror']

    vfs = []
    sec_id = 1
    vhost_id = 0
    for map in dpdk_port_mappings.split(','):
        pci, phys, num_vhost, core_mask = map.split('#')
        num_vhost = int(num_vhost)
        cores = get_cores(core_mask)
        if def_vfs:
            components = def_vfs[sec_id - 1]['components']
        else:
            components = conf_components(sec_id, vhost_id, num_vhost, cores)
        vf = {'physical_network': phys, 'pci_address': pci,
              'num_vhost': num_vhost, 'core_mask': core_mask,
              'components': components}
        vfs.append(vf)
        sec_id += 1
        vhost_id += num_vhost

    host_conf = {'vf': vfs}
    if spp_mirror:
        if def_mirror:
            mirror = def_mirror
        else:
            num_mirror, core_mask = spp_mirror.split('#')
            num_mirror = int(num_mirror)
            cores = get_cores(core_mask)
            mirror = mirror_components(vhost_id, num_mirror, cores)
        host_conf['mirror'] = mirror

    etcd = etcd3.client(etcd_host, etcd_port)
    etcd.put(config_path(host), json.dumps(host_conf))

    num = 0
    for vf in vfs:
        phys = vf['physical_network']
        for i in range(vf['num_vhost']):
            etcd.put(vhost_path(host, phys, num), 'None')
            num += 1

    if spp_mirror:
        for i in range(len(mirror)):
            etcd.put(mirror_path(host, i), 'None')


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2020 NTT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import etcd3
import json
import sys

SPP_ROOT = "/spp/openstack/"


def config_path(host):
    return SPP_ROOT + "configuration/" + host


def vhost_path(host, phys_net, vhost):
    return SPP_ROOT + "vhost/%s/%s/%d" % (host, phys_net, vhost)


def main():
    if len(sys.argv) < 5:
        print("usage: spp-config-build dpdk_port_mappings"
              " host etcd_host etcd_port")
        return 1
    dpdk_port_mappings = sys.argv[1]
    host = sys.argv[2]
    etcd_host = sys.argv[3]
    etcd_port = sys.argv[4]

    confs = json.loads(
        dpdk_port_mappings.replace("u'", "\"").replace("'", "\""))

    etcd = etcd3.client(etcd_host, etcd_port)
    etcd.put(config_path(host), json.dumps(confs))

    num = 0
    for conf in confs:
        phys = conf['physical_network']
        for i in range(int(conf['num_vhost'])):
            etcd.put(vhost_path(host, phys, num), 'None')
            num += 1


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2017 NTT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from networking_spp._i18n import _


etcd_opts = [
    cfg.StrOpt('etcd_host', default='',
               help=_("Etcd host")),
    cfg.IntOpt('etcd_port', default=2379,
               help=_("Etcd port")),
]


cfg.CONF.register_opts(etcd_opts, 'spp')



# Copyright 2017 NTT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests

from oslo_log import log as logging


LOG = logging.getLogger(__name__)


class SppVfApiCommon(object):

    def __init__(self, sec_id, api_ip_addr, api_port, proc_type):
        self.sec_id = sec_id
        self.api_port = api_port
        self.host_url = "http://%s:%d" % (api_ip_addr, api_port)
        self.proc_url = "/v1/%ss/%d" % (proc_type, sec_id)
        self.info = None

    def send_request(self, method, url, data=None):
        path = self.host_url + self.proc_url + url
        r = requests.request(method, path, json=data)
        if r.status_code >= 400:
            raise RuntimeError("%s %s error: %s" %
                               (method, self.proc_url + url, r.text))
        if r.status_code != 204:
            return r.json()

    def get_status(self):
        ret = self.send_request("GET", "")
        LOG.info("info %d: %s", self.sec_id, ret)
        self.info = ret

    def make_component(self, comp_name, core_id, comp_type):
        data = {"name": comp_name, "core": core_id, "type": comp_type}
        self.send_request("POST", "/components", data)

    def port_add(self, port, direction, comp_name, op=None, vlan_id=None):
        data = {"action": "attach", "port": port, "dir": direction}
        if op:
            if op == "add_vlantag":
                vlan = {"operation": "add", "id": vlan_id, "pcp": 0}
            else:
                vlan = {"operation": "del"}
            data["vlan"] = vlan
        path = "/components/%s/ports" % comp_name
        self.send_request("PUT", path, data)

    def port_del(self, port, direction, comp_name):
        data = {"action": "detach", "port": port, "dir": direction}
        path = "/components/%s/ports" % comp_name
        self.send_request("PUT", path, data)

    def port_exist(self, port, direction, comp_name):
        for comp in self.info["components"]:
            if comp["name"] == comp_name:
                ports = (comp["rx_port"] if direction == "rx"
                         else comp["tx_port"])
                for p in ports:
                    if port == p["port"]:
                        return True
        return False


class SppVfApi(SppVfApiCommon):

    def __init__(self, sec_id, api_ip_addr, api_port):
        super(SppVfApi, self).__init__(sec_id, api_ip_addr, api_port, "vf")

    def set_classifier_table(self, mac_address, port):
        data = {"action": "add", "type": "mac", "mac_address": mac_address,
                "port": port}
        self.send_request("PUT", "/classifier_table", data)

    def clear_classifier_table(self, mac_address, port):
        data = {"action": "del", "type": "mac", "mac_address": mac_address,
                "port": port}
        self.send_request("PUT", "/classifier_table", data)

    def set_classifier_table_with_vlan(self, mac_address, port, vlan_id):
        data = {"action": "add", "type": "vlan", "mac_address": mac_address,
                "port": port, "vlan": vlan_id}
        self.send_request("PUT", "/classifier_table", data)

    def clear_classifier_table_with_vlan(self, mac_address, port, vlan_id):
        data = {"action": "del", "type": "vlan", "mac_address": mac_address,
                "port": port, "vlan": vlan_id}
        self.send_request("PUT", "/classifier_table", data)


class SppMirrorApi(SppVfApiCommon):

    def __init__(self, sec_id, api_ip_addr, api_port):
        super(SppMirrorApi, self).__init__(sec_id, api_ip_addr, api_port,
                                           "mirror")

    def make_component(self, comp_name, core_id):
        super(SppMirrorApi, self).make_component(comp_name, core_id,
                                                 "mirror")



# Copyright (c) 2017 NTT
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import netaddr


from neutron_tempest_plugin import config
from neutron_tempest_plugin.scenario import base
from neutron_tempest_plugin.scenario import constants
from tempest.common import waiters
from tempest.lib.common import ssh
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators

CONF = config.CONF


class SppTest(base.BaseTempestTestCase):
    credentials = ['primary', 'admin']
    force_tenant_isolation = False

    # Default to ipv4.
    _ip_version = 4

    @classmethod
    def setup_clients(cls):
        super(SppTest, cls).setup_clients()
        cls.client = cls.os_admin.network_client

    def _test_spp_connect(self, **net_kwargs):
        hypers = self.os_admin.hv_client.list_hypervisors()['hypervisors']
        azs = self.os_admin.az_client.list_availability_zones()
        az = str(azs['availabilityZoneInfo'][0]['zoneName'])
        h1 = az + ':' + str(hypers[0]['hypervisor_hostname'])
        h2 = az + ':' + str(hypers[1]['hypervisor_hostname'])

        network = self.create_network()
        subnet = self.create_subnet(network)

        network1 = self.create_network('network1', **net_kwargs)
        self.create_subnet(network1, reserve_cidr=False,
                           cidr=netaddr.IPNetwork('30.0.0.0/28'))

        secgroup = self.os_admin.network_client.create_security_group(
            name=data_utils.rand_name('secgroup'))
        self.security_groups.append(secgroup['security_group'])
        router = self.create_router_by_client()
        self.create_router_interface(router['id'], subnet['id'])
        clientk = self.os_admin.keypairs_client
        keypair = self.create_keypair(client=clientk)
        self.create_loginable_secgroup_rule(
            secgroup_id=secgroup['security_group']['id'],
            client=self.os_admin.network_client)

        server1 = self.create_server(
            flavor_ref=CONF.compute.flavor_ref,
            image_ref=CONF.compute.image_ref,
            key_name=keypair['name'],
            networks=[{'uuid': network['id']}, {'uuid': network1['id']}],
            security_groups=[{'name': secgroup['security_group']['name']}],
            availability_zone=h1
            )
        server2 = self.create_server(
            flavor_ref=CONF.compute.flavor_ref,
            image_ref=CONF.compute.image_ref,
            key_name=keypair['name'],
            networks=[{'uuid': network['id']}, {'uuid': network1['id']}],
            security_groups=[{'name': secgroup['security_group']['name']}],
            availability_zone=h2
            )

        waiters.wait_for_server_status(self.os_admin.servers_client,
                                       server1['server']['id'],
                                       constants.SERVER_STATUS_ACTIVE)
        waiters.wait_for_server_status(self.os_admin.servers_client,
                                       server2['server']['id'],
                                       constants.SERVER_STATUS_ACTIVE)

        port1 = self.client.list_ports(network_id=network['id'],
                                       device_id=server1[
                                       'server']['id'])['ports'][0]
        port2 = self.client.list_ports(network_id=network['id'],
                                       device_id=server2[
                                       'server']['id'])['ports'][0]

        clientp = self.os_admin.network_client
        fip1 = self.create_and_associate_floatingip(port1['id'],
                                                    client=clientp)

        fip2 = self.create_and_associate_floatingip(port2['id'],
                                                    client=clientp)

        # NOTE: initialization takes a long time after status becomes ACTIVE.
        ssh_client1 = ssh.Client(fip1['floating_ip_address'],
                                 CONF.validation.image_ssh_user,
                                 pkey=keypair['private_key'],
                                 timeout=400)

        ssh_client2 = ssh.Client(fip2['floating_ip_address'],
                                 CONF.validation.image_ssh_user,
                                 pkey=keypair['private_key'])

        ssh_client1.exec_command('dpdk_setup.sh')
        ssh_client2.exec_command('dpdk_setup.sh')

        ssh_client1.exec_command('sudo systemctl start dpdkapp_pong')

        cmd = 'sudo l2_ping_pong -- --ping | tail -n 1'
        res = ssh_client2.exec_command(cmd)
        print(res)

        self.assertTrue('Success.' in str(res))

    @decorators.idempotent_id('1361878f-c6b7-461f-9b24-0c3a8807f1e6')
    def test_spp_connect_flat(self):
        net_kwargs = {'provider:network_type': 'flat',
                      'provider:physical_network': 'phys1'}
        self._test_spp_connect(**net_kwargs)

    @decorators.idempotent_id('fe8471be-4d7b-4d74-9c67-8ba7a5e6859b')
    def test_spp_connect_vlan(self):
        net_kwargs = {'provider:network_type': 'vlan',
                      'provider:physical_network': 'phys2'}
        self._test_spp_connect(**net_kwargs)



# Copyright (c) 2017 NTT
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock

from networking_spp.mech_driver import mech_spp
from neutron.tests import base
from neutron_lib.api.definitions import portbindings
from neutron_lib import constants
from neutron_lib.plugins.ml2 import api


class SppMechanismTestCase(base.BaseTestCase):

    def setUp(self):
        super(SppMechanismTestCase, self).setUp()
        self.driver = mech_spp.SppMechanismDriver()
        self.driver.etcd = mock.Mock()
        self.context = mock.Mock()

    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'SppMechanismDriver._spp_agent_alive')
    def test_bind_port_agent_conf(self, mocked_spp_agent_alive):
        self.driver.etcd.get.return_value = None
        self.driver.bind_port(self.context)
        mocked_spp_agent_alive.assert_not_called()

    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'LOG.warn')
    def test_bind_port_spp_agent_alive(self, mocked_warn):
        agent_conf = {'vf': [{'physical_network': 'phy_net'}]}
        self.driver.etcd.get.return_value = json.dumps(agent_conf)
        self.context.host_agents.return_value = []
        self.driver.bind_port(self.context)
        mocked_warn.assert_called()

    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'SppMechanismDriver._spp_agent_alive')
    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'SppMechanismDriver._try_to_bind')
    def test_bind_port_call_try_to_bind(self, mocked_try_to_bind,
                                        mocked_spp_agent_alive):
        # conditions to call _try_to_bind
        agent_conf = {'vf': [{'physical_network': 'phy_net'}]}
        self.driver.etcd.get.return_value = json.dumps(agent_conf)
        mocked_spp_agent_alive.return_value = True
        segment = {api.NETWORK_TYPE: constants.TYPE_FLAT,
                   api.PHYSICAL_NETWORK: 'phy_net'}
        self.context.segments_to_bind = [segment]
        self.driver.bind_port(self.context)
        mocked_try_to_bind.assert_called()

    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'SppMechanismDriver._spp_agent_alive')
    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'SppMechanismDriver._try_to_bind')
    def test_bind_port_call_try_to_bind_vlan(self, mocked_try_to_bind,
                                             mocked_spp_agent_alive):
        # conditions to call _try_to_bind
        agent_conf = {'vf': [{'physical_network': 'phy_net'}]}
        self.driver.etcd.get.return_value = json.dumps(agent_conf)
        mocked_spp_agent_alive.return_value = True
        segment = {api.NETWORK_TYPE: constants.TYPE_VLAN,
                   api.PHYSICAL_NETWORK: 'phy_net'}
        self.context.segments_to_bind = [segment]
        self.driver.bind_port(self.context)
        mocked_try_to_bind.assert_called()

    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'SppMechanismDriver._spp_agent_alive')
    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'SppMechanismDriver._try_to_bind')
    def test_bind_port_not_call_try_to_bind(self, mocked_try_to_bind,
                                            mocked_spp_agent_alive):
        # conditions not to call _try_to_bind
        agent_conf = {'vf': []}
        self.driver.etcd.get.return_value = json.dumps(agent_conf)
        mocked_spp_agent_alive.return_value = True
        segment = {api.NETWORK_TYPE: 'value', api.PHYSICAL_NETWORK: 'value'}
        self.context.segments_to_bind = [segment]
        self.driver.bind_port(self.context)
        mocked_try_to_bind.assert_not_called()

    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'SppMechanismDriver._unplug_port')
    def test_update_port_postcommit_con1(self, mocked_unplug_port):
        # conditions to call _unplug_port
        self.context.original_host = True
        self.context.original_vif_type = 'vhostuser'
        self.context.host = ''
        self.context.current = {'id': 'value'}
        self.context.vif_type = 'unbound'

        self.driver.update_port_postcommit(self.context)
        mocked_unplug_port.assert_called()

    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'SppMechanismDriver._unplug_port')
    def test_update_port_postcommit_con2(self, mocked_unplug_port):
        # conditions not to call _unplug_port
        self.context.original_host = None
        self.context.original_vif_type = ''
        self.context.host = ''
        self.context.vif_type = 'unbound'

        self.driver.update_port_postcommit(self.context)
        mocked_unplug_port.assert_not_called()

    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'SppMechanismDriver._unplug_port')
    def test_update_port_postcommit_con3(self, mocked_unplug_port):
        # conditions to call _unplug_port
        self.context.original_host = True
        self.context.original_vif_type = 'binding_failed'
        self.context.host = ''
        self.context.current = {'id': 'value'}

        self.driver.update_port_postcommit(self.context)
        mocked_unplug_port.assert_called()

    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'SppMechanismDriver._unplug_port')
    def test_delete_port_postcommit_con1(self, mocked_unplug_port):
        # conditions to call _unplug_port
        self.context.host = 'host'
        self.context.vif_type = 'vhostuser'
        self.context.current = dict(id='value')

        self.driver.delete_port_postcommit(self.context)
        mocked_unplug_port.assert_called()

    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'SppMechanismDriver._unplug_port')
    def test_delete_port_postcommit_con2(self, mocked_unplug_port):
        # conditions not to call _unplug_port
        self.context.host = 'host'
        self.context.vif_type = ''

        self.driver.delete_port_postcommit(self.context)
        mocked_unplug_port.assert_not_called()

    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'SppMechanismDriver._unplug_port')
    def test_delete_port_postcommit_con3(self, mocked_unplug_port):
        # conditions to call _unplug_port
        self.context.host = 'host'
        self.context.vif_type = 'binding_failed'
        self.context.current = dict(id='value')

        self.driver.delete_port_postcommit(self.context)
        mocked_unplug_port.assert_called()

    def _get(self, key):
        if key:
            return key

    def _put(self, key, action):
        if key:
            return action + key

    @mock.patch('networking_spp.common.etcd_key.action_key')
    def test_unplug_port_con1(self, mocked_action_key):

        # conditions to call etcd.put
        mocked_action_key.return_value = 'key'
        self.driver.etcd.get.side_effect = self._get
        self.driver.etcd.put.side_effect = self._put
        self.driver._unplug_port('host', '1')
        self.driver.etcd.put.assert_called()

    @mock.patch('networking_spp.common.etcd_key.action_key')
    def test_unplug_port_con2(self, mocked_action_key):

        # conditions not to call etcd.put
        mocked_action_key.return_value = None
        self.driver.etcd.get.side_effect = self._get
        self.driver.etcd.put.side_effect = self._put
        self.driver._unplug_port('host', '1')
        self.driver.etcd.put.assert_not_called()

    @mock.patch('networking_spp.common.etcd_key.port_status_key')
    @mock.patch('networking_spp.mech_driver.mech_spp.LOG.debug')
    def test_wait_plug_port_con1(self, mocked_debug,
                                 mocked_port_status_key):
        # conditions to return True
        mocked_port_status_key.return_value = 'key'
        mocked_debug.return_value = None
        self.driver.etcd.watch_once.return_value = ('key', 'up')
        value = self.driver._wait_plug_port('host', '1')
        self.assertEqual(value, True)

    @mock.patch('networking_spp.common.etcd_key.port_status_key')
    def test_wait_plug_port_con2(self, mocked_port_status_key):

        # conditions to return None
        mocked_port_status_key.return_value = ''
        self.driver.etcd.watch_once.return_value = (None, None)
        value = self.driver._wait_plug_port('host', '1')
        self.assertEqual(value, None)

    @mock.patch('networking_spp.common.etcd_key.action_key')
    @mock.patch('networking_spp.common.etcd_key.bind_port_key')
    def test_add_bind_port(self, mocked_bind_port_key,
                           mocked_action_key):

        # conditions not to call etcd.put
        self.context.current = {'mac_address': '00:00:00:00:00:00',
                                'id': 'AAAABBBB'}
        self.context.host = 'host1'
        mocked_action_key.side_effect = lambda host, port: host + port
        mocked_bind_port_key.side_effect = lambda host, port: host + port
        self.driver.etcd.put.side_effect = self._put
        self.driver._add_bind_port(self.context, '1', None)
        args, w = self.driver.etcd.put.call_args
        self.assertEqual(args[0], 'host1AAAABBBB')
        self.assertEqual(self.driver.etcd.put.call_count, 2)

    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'SppMechanismDriver._wait_plug_port')
    @mock.patch('networking_spp.mech_driver.mech_spp.'
                'SppMechanismDriver._add_bind_port')
    @mock.patch('networking_spp.common.etcd_key.vhost_phys_prefix')
    def test_try_to_bind(self,
                         mocked_vhost_phys_prefix,
                         mocked_add_bind_port,
                         mocked_wait_plug_port):
        # conditions to find vhost
        self.context.current = {'mac_address': '00:00:00:00:00:00',
                                'id': '5678BBBB'}
        self.context.host = 'host1'
        segment = {api.NETWORK_TYPE: constants.TYPE_FLAT,
                   api.PHYSICAL_NETWORK: 'phy1',
                   api.SEGMENTATION_ID: None,
                   api.ID: 'id'}
        mocked_vhost_phys_prefix.side_effect = \
            lambda a, b: '/x/y/%s/%s/' % (a, b)
        value = [('/x/y/host1/phy1/1', '1234AAAA'),
                 ('/x/y/host1/phy1/2', 'EEEE4444'),
                 ('/x/y/host1/phy1/3', 'None')]
        self.driver.etcd.get.return_value = None
        self.driver.etcd.get_prefix.return_value = value
        self.driver.etcd.replace.return_value = True
        mocked_wait_plug_port.return_value = True
        mocked_add_bind_port.return_value = None

        self.driver._try_to_bind(self.context, segment)
        self.context.set_binding.assert_called()

        args, wagrs = self.context.set_binding.call_args
        vif_type = args[1]
        expected_vif_type = 'vhostuser'
        self.assertEqual(vif_type, expected_vif_type)
        vif_details = args[2]
        sock_path = "/tmp/sock3"
        mode = portbindings.VHOST_USER_MODE_SERVER
        expected_vif_details = {portbindings.CAP_PORT_FILTER: False,
                                portbindings.VHOST_USER_MODE: mode,
                                portbindings.VHOST_USER_SOCKET: sock_path}
        self.assertEqual(vif_details, expected_vif_details)

    @mock.patch('networking_spp.mech_driver.mech_spp.LOG.warn')
    @mock.patch('networking_spp.common.etcd_key.vhost_phys_prefix')
    def test_try_to_bind_not_vhost(self, mocked_vhost_phys_prefix,
                                   mocked_warn):
        # conditions not to find vhost
        self.context.current = {'mac_address': '00:00:00:00:00:00',
                                'id': '5678BBBB'}
        self.context.host = 'host1'
        segment = {api.NETWORK_TYPE: constants.TYPE_FLAT,
                   api.PHYSICAL_NETWORK: 'phy1',
                   api.ID: 'id'}
        mocked_vhost_phys_prefix.side_effect = \
            lambda a, b: '/x/y/%s/%s/' % (a, b)
        value = [('/x/y/host1/phy1/1', '1234AAAA'),
                 ('/x/y/host1/phy1/2', 'EEEE4444'),
                 ('/x/y/host1/phy1/3', '99999999')]
        self.driver.etcd.get.return_value = None
        self.driver.etcd.get_prefix.return_value = value
        self.driver.etcd.replace.return_value = True

        self.driver._try_to_bind(self.context, segment)
        mocked_warn.assert_called()



# Copyright 2017 NTT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from networking_spp._i18n import _
from neutron.conf.agent import common


spp_opts = [
    cfg.StrOpt('api_ip_addr', default='127.0.0.1',
               help=_("Spp-ctl bind ip address")),
    cfg.IntOpt('api_port', default=7777,
               help=_("Spp-ctl web api port number")),
]


cfg.CONF.register_opts(spp_opts, 'spp')
common.register_agent_state_opts_helper(cfg.CONF)



# Copyright (c) 2018 NTT
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from networking_spp.common import etcd_client
from networking_spp.common import etcd_config  # noqa
from networking_spp.common import etcd_key
from neutron_lib import exceptions as n_exc
from neutron_taas.services.taas import service_drivers as taas_service_drivers
from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log


LOG = log.getLogger(__name__)

WAIT_TAP_TIMEOUT = 30


class PortNotBound(n_exc.BadRequest):
    message = _("Port %(port_id)s is not bound on SPP network host")


class NotSameHost(n_exc.BadRequest):
    message = _("Source port's host must be same as Service port's host")


class NoMirrorAvailable(n_exc.Conflict):
    message = _("No mirror available on the host %(host)s")


class SppTaasDriver(taas_service_drivers.TaasBaseDriver):

    def __init__(self, service_plugin):
        LOG.debug("Loading SppTaasDriver.")
        super(SppTaasDriver, self).__init__(service_plugin)
        self.etcd = etcd_client.EtcdClient(cfg.CONF.spp.etcd_host,
                                           cfg.CONF.spp.etcd_port)

    @log_helpers.log_method_call
    def create_tap_service_precommit(self, context):
        # nothing to do
        pass

    def _check_bind_port(self, host, port_id):
        if (host is None or
                self.etcd.get(etcd_key.bind_port_key(host, port_id)) is None):
            raise PortNotBound(port_id=port_id)

    @log_helpers.log_method_call
    def create_tap_service_postcommit(self, context):
        # check port bound to SPP
        port_id = context.tap_service['port_id']
        port = self.service_plugin._get_port_details(context._plugin_context,
                                                     port_id)
        self._check_bind_port(port['binding:host_id'], port_id)
        # that's all. check only at the moment.

    @log_helpers.log_method_call
    def delete_tap_service_precommit(self, context):
        # nothing to do
        pass

    @log_helpers.log_method_call
    def delete_tap_service_postcommit(self, context):
        # nothing to do
        pass

    @log_helpers.log_method_call
    def create_tap_flow_precommit(self, context):
        # nothing to do
        pass

    def _get_mirror(self, host, tf_id):
        prefix = etcd_key.mirror_prefix(host)
        mirror_id = None
        for key, value in self.etcd.get_prefix(prefix):
            if value == 'None':
                if self.etcd.replace(key, 'None', tf_id):
                    mirror_id = key[len(prefix):]
                    LOG.debug("mirror %s assigned for tap_flow %s",
                              mirror_id, tf_id)
                    break
        return mirror_id

    def _free_mirror(self, host, mirror_id):
        self.etcd.put(etcd_key.mirror_key(host, mirror_id), 'None')

    @log_helpers.log_method_call
    def create_tap_flow_postcommit(self, context):
        # check source port bound to SPP
        tf = context.tap_flow
        src_port_id = tf['source_port']
        src_port = self.service_plugin._get_port_details(
            context._plugin_context, src_port_id)
        host = src_port['binding:host_id']
        self._check_bind_port(host, src_port_id)

        # check service port bound to SPP on the same host
        ts = self.service_plugin.get_tap_service(context._plugin_context,
                                                 tf['tap_service_id'])
        svc_port_id = ts['port_id']
        if self.etcd.get(etcd_key.bind_port_key(host, svc_port_id)) is None:
            raise NotSameHost()

        # assign mirror(s)
        mirror_in = None
        mirror_out = None
        tf_id = tf['id']
        direction = tf['direction']
        if direction == 'IN' or direction == 'BOTH':
            mirror_in = self._get_mirror(host, tf_id)
            if mirror_in is None:
                raise NoMirrorAvailable(host=host)
        if direction == 'OUT' or direction == 'BOTH':
            mirror_out = self._get_mirror(host, tf_id)
            if mirror_out is None:
                if mirror_in:
                    self._free_mirror(host, mirror_in)
                raise NoMirrorAvailable(host=host)

        # request tap setup
        value = {"service_port": svc_port_id,
                 "source_port": src_port_id,
                 "mirror_in": int(mirror_in) if mirror_in else None,
                 "mirror_out": int(mirror_out) if mirror_out else None}
        key = etcd_key.tap_info_key(host, tf_id)
        self.etcd.put(key, json.dumps(value))
        key = etcd_key.tap_action_key(host, tf_id)
        self.etcd.put(key, "plug")

        # wait tap setup done
        key = etcd_key.tap_status_key(host, tf_id)
        k, v = self.etcd.watch_once(key, timeout=WAIT_TAP_TIMEOUT)
        if k and v == "up":
            LOG.debug("tap_flow %s plug done.", tf_id)
        else:
            LOG.warning("tap_flow %s plug not complete.", tf_id)
            # NOTE: if it only takes time, it may not be a problem.
            # but if there is agent's problem, user is only availble
            # to delete the tap_flow.

    @log_helpers.log_method_call
    def delete_tap_flow_precommit(self, context):
        # nothing to do
        pass

    @log_helpers.log_method_call
    def delete_tap_flow_postcommit(self, context):
        tf = context.tap_flow
        src_port_id = tf['source_port']
        host = None
        try:
            src_port = self.service_plugin._get_port_details(
                context._plugin_context, src_port_id)
            host = src_port['binding:host_id']
        except n_exc.PortNotFound:
            pass
        if host is None:
            # port is already deleted or unbound. in this case,
            # tap unplug was already done by agent. so nothing to do.
            return

        key = etcd_key.tap_action_key(host, tf['id'])
        if self.etcd.get(key) is not None:
            # NOTE: if service port is deleted or unbound, tap
            # unplug was done by agent. in this case, the key
            # has been deleted already. if key exists, request
            # unplug.
            self.etcd.put(key, "unplug")



# Copyright (c) 2017 NTT
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock

from networking_spp.agent import spp_agent
from neutron.tests import base


class SppVfTestCase(base.BaseTestCase):

    def setUp(self):
        super(SppVfTestCase, self).setUp()
        self.spp_vf = spp_agent.SppVf(1, 'phy_net', '127.0.0.1', 7777)

    @mock.patch('networking_spp.agent.spp_api.SppVfApi.make_component')
    @mock.patch('networking_spp.agent.spp_api.SppVfApi.port_add')
    @mock.patch('networking_spp.agent.spp_api.SppVfApi.get_status')
    def test_build_components_con1(self, m_g, m_port_add, m_mk_comp):
        self.spp_vf.info = {"components": [{"core": 2,
                                            "type": "forward",
                                            "name": "fw1",
                                            "rx_port": [{"port": "vhost:2"}],
                                            "tx_port": [{"port": "ring:2"}]}]}
        components = [{"core": 2,
                       "type": "forward",
                       "name": "fw2",
                       "rx_port": ["vhost:1"],
                       "tx_port": ["ring:1"]}]
        self.spp_vf.build_components(components)
        self.assertEqual(m_mk_comp.call_count, 1)
        self.assertEqual(m_port_add.call_count, 2)
        m_port_add.assert_called_with("vhost:1", "rx", "fw2")

    @mock.patch('networking_spp.agent.spp_api.SppVfApi.make_component')
    @mock.patch('networking_spp.agent.spp_api.SppVfApi.port_add')
    @mock.patch('networking_spp.agent.spp_api.SppVfApi.get_status')
    def test_build_components_con2(self, m_g, m_port_add, m_mk_comp):
        self.spp_vf.info = {"components": [{"core": 2,
                                            "type": "classifier",
                                            "name": "fw3",
                                            "rx_port": [],
                                            "tx_port": [{"port": "ring:1"}]}]}
        components = [{"core": 2,
                       "type": "classifier",
                       "name": "fw3",
                       "rx_port": ["vhost:2"],
                       "tx_port": ["ring:1", "ring:2", "ring:3"]}]
        self.spp_vf.build_components(components)
        self.assertEqual(m_mk_comp.call_count, 0)
        self.assertEqual(m_port_add.call_count, 3)
        m_port_add.assert_called_with("vhost:2", "rx", "fw3")

    @mock.patch('networking_spp.agent.spp_api.SppVfApi.make_component')
    @mock.patch('networking_spp.agent.spp_api.SppVfApi.port_add')
    @mock.patch('networking_spp.agent.spp_api.SppVfApi.get_status')
    def test_build_components_con3(self, m_g, m_port_add, m_mk_comp):
        self.spp_vf.info = {"components": [{"core": 2,
                                            "type": "forward",
                                            "name": "fw4",
                                            "rx_port": [],
                                            "tx_port": []}]}
        components = [{"core": 2,
                       "type": "forward",
                       "name": "fw4",
                       "rx_port": ["vhost:3"],
                       "tx_port": ["ring:3"]}]
        self.spp_vf.build_components(components)
        self.assertEqual(m_mk_comp.call_count, 0)
        self.assertEqual(m_port_add.call_count, 2)
        m_port_add.assert_called_with("vhost:3", "rx", "fw4")

    @mock.patch('networking_spp.agent.spp_api.SppVfApi.make_component')
    @mock.patch('networking_spp.agent.spp_api.SppVfApi.port_add')
    @mock.patch('networking_spp.agent.spp_api.SppVfApi.get_status')
    def test_build_components_con4(self, m_g, m_port_add, m_mk_comp):
        self.spp_vf.info = {"components": [{"core": 2,
                                            "type": "forward",
                                            "name": "fw5",
                                            "rx_port": [],
                                            "tx_port": []}]}
        components = [{"core": 2,
                       "type": "merge",
                       "name": "fw5",
                       "rx_port": ["ring:1"],
                       "tx_port": ["vhost:1"]}]
        self.spp_vf.build_components(components)
        self.assertEqual(m_mk_comp.call_count, 0)
        self.assertEqual(m_port_add.call_count, 2)
        m_port_add.assert_called_with("ring:1", "rx", "fw5")

    @mock.patch('networking_spp.agent.spp_api.SppVfApi.make_component')
    @mock.patch('networking_spp.agent.spp_api.SppVfApi.port_add')
    @mock.patch('networking_spp.agent.spp_api.SppVfApi.get_status')
    def test_build_components_con5(self, m_g, m_port_add, m_mk_comp):
        self.spp_vf.info = {"components": [{"core": 2,
                                            "type": "forward",
                                            "name": "fw6",
                                            "rx_port": [{"port": "phys:2"}],
                                            "tx_port": []}]}
        components = [{"core": 2,
                       "type": "merge",
                       "name": "fw6",
                       "rx_port": ["phys:1", "phys:3", "phys:2"],
                       "tx_port": ["phys:5"]}]
        self.spp_vf.build_components(components)
        self.assertEqual(m_mk_comp.call_count, 0)
        self.assertEqual(m_port_add.call_count, 3)
        m_port_add.assert_called_with("phys:3", "rx", "fw6")

    def test_build_vhosts_con1(self):
        components = [{"type": "forward",
                       "name": "fw0",
                       "rx_port": ["ring:0"],
                       "tx_port": ["vhost:0"]},
                      {"type": "forward",
                       "name": "fw1",
                       "rx_port": ["vhost:0"],
                       "tx_port": ["ring:1"]},
                      {"type": "classifier",
                       "name": "cls1",
                       "rx_port": ["phy:0"],
                       "tx_port": ["ring:0"]},
                      {"type": "merge",
                       "name": "merge1",
                       "rx_port": ["ring:1"],
                       "tx_port": ["phy:0"]}]
        self.spp_vf.build_vhosts(components)

        vhost = self.spp_vf.vhostusers[0]
        self.assertEqual(vhost.name, "vhost:0")
        self.assertEqual(vhost.del_vlan_port, "ring:0")
        self.assertEqual(vhost.del_vlan_comp, "cls1")
        self.assertEqual(vhost.add_vlan_port, "ring:1")
        self.assertEqual(vhost.add_vlan_comp, "merge1")
        self.assertEqual(vhost.dst_comp, "fw0")
        self.assertEqual(vhost.in_ring, "ring:0")
        self.assertEqual(vhost.in_comp, "fw0")
        self.assertEqual(vhost.out_ring, "ring:1")
        self.assertEqual(vhost.out_comp, "fw1")

    def test_build_vhosts_con2(self):
        components = [{"type": "classifier",
                       "name": "cls2",
                       "rx_port": ["phy:1"],
                       "tx_port": ["vhost:1"]},
                      {"type": "merge",
                       "name": "merge2",
                       "rx_port": ["vhost:1"],
                       "tx_port": ["phy:1"]}]
        self.spp_vf.build_vhosts(components)
        vhost = self.spp_vf.vhostusers[1]
        self.assertEqual(vhost.name, "vhost:1")
        self.assertEqual(vhost.del_vlan_port, "vhost:1")
        self.assertEqual(vhost.del_vlan_comp, "cls2")
        self.assertEqual(vhost.add_vlan_port, "vhost:1")
        self.assertEqual(vhost.add_vlan_comp, "merge2")
        self.assertEqual(vhost.dst_comp, None)
        self.assertEqual(vhost.in_ring, None)
        self.assertEqual(vhost.in_comp, None)
        self.assertEqual(vhost.out_ring, None)
        self.assertEqual(vhost.out_comp, None)

    def test_init_vhost_mac_address_con1(self):
        info = {"classifier_table": [{"type": "mac",
                                      "port": "ring:12",
                                      "value": "12:ab:34:cd:56:ef"}]}
        self.spp_vf.info = info
        self.spp_vf.vhostusers[0] = mock.Mock()
        self.spp_vf.vhostusers[0].del_vlan_port = "ring:12"
        self.spp_vf.init_vhost_mac_address()

        self.assertEqual(self.spp_vf.vhostusers[0].mac_address,
                         '12:ab:34:cd:56:ef')

    def test_init_vhost_mac_address_con2(self):
        info = {"classifier_table": [{"type": "vlan",
                                      "port": "ring:12",
                                      "value": "100/12:ab:34:cd:56:ef"}]}
        self.spp_vf.info = info
        self.spp_vf.vhostusers[0] = mock.Mock()
        self.spp_vf.vhostusers[0].del_vlan_port = "ring:12"
        self.spp_vf.init_vhost_mac_address()

        self.assertEqual(self.spp_vf.vhostusers[0].mac_address,
                         '12:ab:34:cd:56:ef')


class SppMirrorTestCase(base.BaseTestCase):

    def setUp(self):
        super(SppMirrorTestCase, self).setUp()
        self.spp_mirror = spp_agent.SppMirror(1, '127.0.0.1', 7778)

    @mock.patch('networking_spp.agent.spp_api.SppVfApiCommon.get_status')
    @mock.patch('networking_spp.agent.spp_api.SppVfApiCommon.make_component')
    def test_init_components(self, m_mk_comp, m_get):
        self.spp_mirror.info = {"components": [{"core": 2,
                                                "type": "forward",
                                                "name": "fw1",
                                                "ports": [7777, 7778]},
                                               {"core": 2,
                                                "type": "unuse",
                                                "name": "fw3",
                                                "ports": [7777, 7778]},
                                               {"core": 2,
                                                "type": "forward",
                                                "name": "fw4",
                                                "ports": [7777, 7778]}
                                               ]}
        components = [{"core": 3,
                       "type": "forward",
                       "name": "fw5",
                       "ports": [7777, 7778]}]

        self.spp_mirror.init_components(components)
        mirror = spp_agent.Mirror("mirror_0", [7777, 7778], self.spp_mirror)
        self.assertEqual(m_mk_comp.call_count, 1)
        m_mk_comp.assert_called_with('mirror_0', 3, 'mirror')
        self.assertEqual(len(self.spp_mirror.mirrors), 1)
        self.assertEqual(self.spp_mirror.mirrors[0].comp, mirror.comp)
        self.assertEqual(self.spp_mirror.mirrors[0].proc, mirror.proc)
        self.assertEqual(self.spp_mirror.mirrors[0].ring_b, mirror.ring_b)
        self.assertEqual(self.spp_mirror.mirrors[0].ring_b, mirror.ring_b)


class SppAgentTestCase(base.BaseTestCase):

    def setUp(self):
        super(SppAgentTestCase, self).setUp()
        self.sec_id = 1
        self.vhost_id = 2
        self.conf = mock.Mock()
        self.agent = self._get_agent()

    @mock.patch('networking_spp.agent.spp_agent.SppMirror')
    @mock.patch('networking_spp.agent.spp_agent.SppAgent.'
                'get_spp_configuration')
    @mock.patch('networking_spp.agent.spp_agent.SppVf')
    @mock.patch('networking_spp.common.etcd_client.EtcdClient')
    @mock.patch('networking_spp.agent.spp_agent.SppAgent._report_state')
    @mock.patch('networking_spp.agent.spp_agent.eventlet')
    def _get_agent(self, a, b, c, d, e, f):
        return spp_agent.SppAgent(self.conf)

    def test_get_spp_configuration(self):
        # conditions
        self.agent.host = 'host1'
        key = '/spp/openstack/configuration/' + self.agent.host
        value = {key: '1'}
        self.agent.etcd.get.return_value = json.dumps(value)
        ret = self.agent.get_spp_configuration()

        self.agent.etcd.get.assert_called_with(key)
        self.assertEqual(ret, value)

    def test_set_classifier_table_con1(self):
        # conditions to exec if statement
        mac = '12:ab:34:cd:56:ef'
        vhost = mock.Mock()
        vhost.vf = mock.Mock()
        vhost.mac_address = mac
        self.agent.vhostusers[1] = vhost

        ret = self.agent.set_classifier_table(1, mac, None)
        self.assertEqual(ret, None)
        self.assertEqual(vhost.vf.set_classifier_table.call_count,
                         0)

    def test_set_classifier_table_con2(self):
        # conditions to exec correctly
        mac = '12:ab:34:cd:56:ef'
        vhost = mock.Mock()
        vhost.vf = mock.Mock()
        vhost.mac_address = 'unuse'
        self.agent.vhostusers[1] = vhost

        self.agent.set_classifier_table(1, mac, None)
        self.assertEqual(vhost.vf.set_classifier_table.call_count,
                         1)
        self.assertEqual(vhost.mac_address, mac)

    def test_set_classifier_table_with_vlan(self):
        mac = '12:ab:34:cd:56:ef'
        vhost = mock.Mock()
        vhost.vf = mock.Mock()
        vhost.mac_address = 'unuse'
        self.agent.vhostusers[1] = vhost

        self.agent.set_classifier_table(1, mac, 10)
        self.assertEqual(vhost.vf.port_add.call_count, 2)
        self.assertEqual(vhost.vf.port_del.call_count, 2)
        self.assertEqual(vhost.vf.set_classifier_table_with_vlan.call_count, 1)
        self.assertEqual(vhost.mac_address, mac)

    def test_clear_classifier_table_con1(self):
        # conditions to exec if statement
        mac = 'unuse'
        vhost = mock.Mock()
        vhost.vf = mock.Mock()
        vhost.mac_address = 'unuse'
        self.agent.vhostusers[1] = vhost

        ret = self.agent.clear_classifier_table(1, mac, None)
        self.assertEqual(ret, None)
        count = vhost.vf.clear_classifier_table.call_count
        self.assertEqual(count, 0)

    def test_clear_classifier_table_con2(self):
        # conditions to exec correctly
        mac = '12:ab:34:cd:56:ef'
        vhost = mock.Mock()
        vhost.vf = mock.Mock()
        vhost.mac_address = mac
        self.agent.vhostusers[1] = vhost

        self.agent.clear_classifier_table(1, mac, None)
        count = vhost.vf.clear_classifier_table.call_count
        self.assertEqual(count, 1)
        self.assertEqual(vhost.mac_address, 'unuse')

    def test_clear_classifier_table_with_vlan(self):
        mac = '12:ab:34:cd:56:ef'
        vhost = mock.Mock()
        vhost.vf = mock.Mock()
        vhost.mac_address = mac
        self.agent.vhostusers[1] = vhost

        self.agent.clear_classifier_table(1, mac, 10)
        self.assertEqual(vhost.vf.port_add.call_count, 2)
        self.assertEqual(vhost.vf.port_del.call_count, 2)
        cnt = vhost.vf.clear_classifier_table_with_vlan.call_count
        self.assertEqual(cnt, 1)
        self.assertEqual(vhost.mac_address, 'unuse')

    @mock.patch('networking_spp.agent.spp_agent.SppAgent.set_classifier_table')
    def test_plug_port(self, m_set):
        # conditions
        port_id = 111
        vhost_id = 1
        mac = '12:ab:34:cd:56:ef'
        self.agent.host = 'host1'
        key = '/spp/openstack/port_status/host1/111'

        self.agent._plug_port(port_id, vhost_id, mac, None)
        m_set.assert_called_with(vhost_id, mac, None)
        self.agent.etcd.put.assert_called_with(key, "up")

    @mock.patch('networking_spp.agent.spp_agent.SppAgent._unplug_tap_port')
    @mock.patch('networking_spp.common.etcd_key.bind_port_key')
    @mock.patch('networking_spp.common.etcd_key.action_key')
    @mock.patch('networking_spp.common.etcd_key.port_status_key')
    @mock.patch('networking_spp.common.etcd_key.vhost_key')
    @mock.patch('networking_spp.agent.spp_agent.SppAgent.'
                'clear_classifier_table')
    def test_unplug_port(self,
                         m_clear,
                         m_vhost_key,
                         m_port_status_key,
                         m_action_key,
                         m_bind_port_key,
                         m_unplug_tap_port):

        # conditions
        port_id = 111
        vhost_id = 1
        mac = '12:ab:34:cd:56:ef'
        self.agent.host = 'host'
        vhostuser = mock.Mock()
        vhostuser.phys_net = 'phy_net'
        self.agent.vhostusers = {1: vhostuser}
        m_vhost_key.return_value = 'key'
        self.agent.mirror = True

        self.agent._unplug_port(port_id, vhost_id, mac, None)
        m_unplug_tap_port.assert_called_with(port_id)
        m_clear.assert_called_with(vhost_id, mac, None)
        m_vhost_key.assert_called_with('host', 'phy_net', vhost_id)
        self.agent.etcd.replace.assert_called_once()
        m_bind_port_key.assert_called_with('host', port_id)
        m_action_key.assert_called_with('host', port_id)
        m_port_status_key.assert_called_with('host', port_id)
        self.assertEqual(self.agent.etcd.delete.call_count, 3)

    def test_do_plug_unplug_con1(self):
        # conditions to exec the 1st if statement
        port_id = 111
        self.agent.etcd.get.return_value = None

        ret = self.agent._do_plug_unplug(port_id)
        self.assertEqual(ret, None)
        self.assertEqual(self.agent.etcd.get.call_count, 1)

    @mock.patch('networking_spp.agent.spp_agent.SppAgent._plug_port')
    def test_do_plug_unplug_con2(self, m_plug_port):
        # conditions to exec the 2nd if statement
        port_id = 111
        data = {"vhost_id": 1, 'mac_address': '12:34:56:78:90:ab'}
        value = json.dumps(data)
        self.agent.etcd.get.side_effect = [value, 'plug']

        self.agent._do_plug_unplug(port_id)
        m_plug_port.assert_called_with(port_id, 1, '12:34:56:78:90:ab', None)

    @mock.patch('networking_spp.agent.spp_agent.SppAgent._unplug_port')
    def test_do_plug_unplug_con3(self, m_unplug_port):
        # conditions to exec the else statement
        port_id = 111
        data = {"vhost_id": 1, 'mac_address': '12:34:56:78:90:ab',
                'vlan_id': 10}
        value = json.dumps(data)
        self.agent.etcd.get.side_effect = [value, 'unplug']

        self.agent._do_plug_unplug(port_id)
        m_unplug_port.assert_called_with(port_id, 1, '12:34:56:78:90:ab', 10)

    @mock.patch('networking_spp.common.etcd_key.action_host_prefix')
    @mock.patch('networking_spp.agent.spp_agent.SppAgent._do_plug_unplug')
    def test_port_plug_watch_con1(self, m_do, m_prefix):
        # conditions to exec if statement and for statement twice
        m_prefix.return_value = '/aa/bb/'
        self.agent.etcd.watch_prefix.return_value = [('/aa/bb/123', 'plug'),
                                                     ('/aa/bb/321', 'unplug')
                                                     ]
        self.agent.port_plug_watch()
        self.assertEqual(m_do.call_args_list[0][0][0], '123')
        self.assertEqual(m_do.call_args_list[1][0][0], '321')
        self.assertEqual(m_do.call_count, 2)

    @mock.patch('networking_spp.common.etcd_key.action_host_prefix')
    @mock.patch('networking_spp.agent.spp_agent.SppAgent._do_plug_unplug')
    def test_port_plug_watch_con2(self, m_do, m_prefix):
        # conditions to exec if statement and then raise exception
        m_prefix.return_value = '/aa/bb/'
        self.agent.etcd.watch_prefix.return_value = [('/aa/bb/123', 'plug'),
                                                     ('/aa/bb/321', 'unplug')
                                                     ]
        m_do.side_effect = Exception
        self.agent.port_plug_watch()
        self.assertEqual(m_do.call_count, 1)
        self.assertEqual(self.agent.port_plug_watch_failed, True)

    @mock.patch('networking_spp.agent.spp_agent.SppAgent._do_tap_plug_unplug')
    @mock.patch('networking_spp.common.etcd_key.action_host_prefix')
    @mock.patch('networking_spp.agent.spp_agent.SppAgent._do_plug_unplug')
    def test_recover(self, m_do, m_prefix, m_tap):
        # conditions to exec if statement and for statement twice
        m_prefix.return_value = '/aa/bb/'
        self.agent.etcd.get_prefix.return_value = [('/aa/bb/123', 'plug'),
                                                   ('/aa/bb/321', 'unplug')
                                                   ]
        self.agent.recover()
        self.assertEqual(m_do.call_args_list[0][0][0], '123')
        self.assertEqual(m_do.call_args_list[1][0][0], '321')
        self.assertEqual(m_do.call_count, 2)
        self.assertEqual(m_tap.call_count, 2)

    def test_handle_signal(self):
        self.agent.shutdown_sem = mock.Mock()
        self.agent._handle_signal('si', 'fr')
        self.agent.shutdown_sem.release.assert_called_once()

    def test_wait_shutdown_con1(self):
        self.agent.port_plug_watch_failed = True
        ret = self.agent.wait_shutdown()
        self.assertEqual(ret, 1)

    def test_wait_shutdown_con2(self):
        self.agent.port_plug_watch_failed = False
        ret = self.agent.wait_shutdown()
        self.assertEqual(ret, 0)

    @mock.patch('networking_spp.agent.spp_agent.loopingcall')
    def test_start_report(self, m_loop):
        heartbeat = mock.Mock()
        heartbeat.start.return_value = None
        m_loop.FixedIntervalLoopingCall.return_value = heartbeat
        self.agent.start_report()
        heartbeat.start.assert_called_once()

    def test_start_state(self):
        # conditions
        self.agent.state_rpc = mock.Mock()
        self.agent.state_rpc.report_state.return_value = None
        self.agent.agent_state = {'host': 'host1', 'start_flag': True}
        self.agent._report_state()

        self.agent.state_rpc.report_state.assert_called_once()
        self.assertEqual(self.agent.agent_state, {'host': 'host1'})

    @mock.patch('networking_spp.common.etcd_key.bind_port_key')
    def test_port_id_to_vhost(self, m_bind_port_key):
        port_id = 111
        value = {'vhost_id': 0}
        self.agent.vhostusers = {0: 'vhost'}
        self.agent.etcd.get.return_value = json.dumps(value)
        self.assertEqual(self.agent._port_id_to_vhost(port_id), 'vhost')

    def test_ring_add(self):
        proc = mock.Mock()
        proc.port_exist.return_value = False
        self.agent._ring_add(proc, 'ring', 'direction', 'comp')
        proc.port_add.assert_called_with('ring', 'direction', 'comp')

    def test_ring_del(self):
        proc = mock.Mock()
        proc.port_exist.return_value = True
        self.agent._ring_del(proc, 'ring', 'direction', 'comp')
        proc.port_del.assert_called_with('ring', 'direction', 'comp')

    def test_attach_ring(self):
        self.agent._ring_add = mock.Mock()
        self.agent._attach_ring('r', 'rx_p', 'rx_c', 'tx_p', 'tx_c')
        self.assertEqual(self.agent._ring_add.call_count, 2)

    def test_detach_ring(self):
        self.agent._ring_del = mock.Mock()
        self.agent._detach_ring('r', 'rx_p', 'rx_c', 'tx_p', 'tx_c')
        self.assertEqual(self.agent._ring_del.call_count, 2)

    def test_change_connection(self):
        self.agent._ring_add = mock.Mock()
        self.agent._ring_del = mock.Mock()
        self.agent._change_connection('r', 'dir', 'd_p', 'd_c', 'a_p', 'a_c')
        self.agent._ring_del.assert_called_with('d_p', 'r', 'dir', 'd_c')
        self.agent._ring_add.assert_called_with('a_p', 'r', 'dir', 'a_c')

    def test_construct_tap_in(self):
        self.agent._attach_ring = mock.Mock()
        self.agent._change_connection = mock.Mock()
        mirror = mock.Mock()
        d_vhost = mock.Mock()
        s_vhost = mock.Mock()
        self.agent._construct_tap_in(mirror, d_vhost, s_vhost)
        self.assertEqual(self.agent._attach_ring.call_count, 2)
        self.assertEqual(self.agent._change_connection.call_count, 1)

    def test_destruct_tap_in(self):
        self.agent._detach_ring = mock.Mock()
        self.agent._change_connection = mock.Mock()
        mirror = mock.Mock()
        d_vhost = mock.Mock()
        s_vhost = mock.Mock()
        self.agent._destruct_tap_in(mirror, d_vhost, s_vhost)
        self.assertEqual(self.agent._detach_ring.call_count, 2)
        self.assertEqual(self.agent._change_connection.call_count, 1)

    def test_construct_tap_out(self):
        self.agent._attach_ring = mock.Mock()
        self.agent._change_connection = mock.Mock()
        mirror = mock.Mock()
        d_vhost = mock.Mock()
        s_vhost = mock.Mock()
        self.agent._construct_tap_out(mirror, d_vhost, s_vhost)
        self.assertEqual(self.agent._attach_ring.call_count, 2)
        self.assertEqual(self.agent._change_connection.call_count, 1)

    def test_destruct_tap_out(self):
        self.agent._detach_ring = mock.Mock()
        self.agent._change_connection = mock.Mock()
        mirror = mock.Mock()
        d_vhost = mock.Mock()
        s_vhost = mock.Mock()
        self.agent._destruct_tap_out(mirror, d_vhost, s_vhost)
        self.assertEqual(self.agent._detach_ring.call_count, 2)
        self.assertEqual(self.agent._change_connection.call_count, 1)

    @mock.patch('networking_spp.common.etcd_key.tap_status_key')
    def test_plug_tap(self, m_tap_status_key):
        self.agent._construct_tap_in = mock.Mock()
        self.agent._construct_tap_out = mock.Mock()
        mirror_in = mock.Mock()
        mirror_out = mock.Mock()
        m_tap_status_key.return_value = 'key'
        self.agent.etcd.put = mock.Mock()
        self.agent.host = 'host'
        self.agent._plug_tap(1, mirror_in, mirror_out, 'd_vhost', 's_vhost')
        self.assertEqual(self.agent._construct_tap_in.call_count, 1)
        self.assertEqual(self.agent._construct_tap_out.call_count, 1)
        m_tap_status_key.assert_called_with('host', 1)
        self.agent.etcd.put.assert_called_with('key', 'up')

    @mock.patch('networking_spp.common.etcd_key.mirror_key')
    @mock.patch('networking_spp.common.etcd_key.tap_info_key')
    @mock.patch('networking_spp.common.etcd_key.tap_action_key')
    @mock.patch('networking_spp.common.etcd_key.tap_status_key')
    def test_unplug_tap(self,
                        m_tap_status_key,
                        m_tap_action_key,
                        m_tap_info_key,
                        m_mirror_key):
        self.agent._destruct_tap_in = mock.Mock()
        self.agent._destruct_tap_out = mock.Mock()
        mirror_in = mock.Mock()
        mirror_out = mock.Mock()
        m_tap_status_key.return_value = 'key1'
        m_tap_action_key.return_value = 'key2'
        m_tap_info_key.return_value = 'key3'
        m_mirror_key.return_value = 'key4'
        keys = ['key3', 'key2', 'key1']
        self.agent.etcd.put = mock.Mock()
        self.agent.etcd.delete = mock.Mock()
        self.agent.host = 'host'
        self.agent._unplug_tap(1, mirror_in, mirror_out, 'd_vhost', 's_vhost')
        self.assertEqual(self.agent._destruct_tap_in.call_count, 1)
        self.assertEqual(self.agent._destruct_tap_out.call_count, 1)

        for i in range(2):
            args = self.agent.etcd.put.call_args_list[i]
            self.assertEqual(args[0], ('key4', 'None'))

        for i in range(3):
            args, kwargs = self.agent.etcd.delete.call_args_list[i]
            self.assertEqual(args[0], keys[i])

    @mock.patch('networking_spp.common.etcd_key.tap_info_key')
    def test_do_tap_plug_unplug_con1(self, m_tap_info_key):
        # conditions to exec the 1st if statement and make it return None
        tap_flow_id = 3
        self.agent.etcd.get.return_value = None
        self.agent._port_id_to_vhost = mock.Mock()
        ret = self.agent._do_tap_plug_unplug(tap_flow_id)
        self.assertEqual(ret, None)
        self.agent._port_id_to_vhost.assert_not_called()

    @mock.patch('networking_spp.common.etcd_key.tap_info_key')
    def test_do_tap_plug_unplug_con2(self, m_tap_info_key):
        # conditions to exec the 2nd if statement and make it return None
        tap_flow_id = 3
        data = {"mirror_in": 'in',
                "mirror_out": 'out',
                "service_port": 'service',
                "source_port": 'source'}
        value = json.dumps(data)
        self.agent.etcd.get.return_value = value
        self.agent._port_id_to_vhost = mock.Mock()
        self.agent.mirror.get_status = mock.Mock()
        self.agent._port_id_to_vhost.side_effect = [None, 'source']
        ret = self.agent._do_tap_plug_unplug(tap_flow_id)
        var1 = self.agent._port_id_to_vhost.call_args_list[0][0][0]
        var2 = self.agent._port_id_to_vhost.call_args_list[1][0][0]
        self.assertEqual(var1, u'service')
        self.assertEqual(var2, u'source')
        self.assertEqual(ret, None)
        self.agent.mirror.get_stauts.assert_not_called()

    @mock.patch('networking_spp.common.etcd_key.tap_info_key')
    def test_do_tap_plug_unplug_con3(self, m_tap_info_key):
        # conditions to exec the 3rd and 4th if statement
        tap_flow_id = 3
        data = {"mirror_in": 'in',
                "mirror_out": 'out',
                "service_port": 'service',
                "source_port": 'source'}
        value = json.dumps(data)
        self.agent.etcd.get.side_effect = [value, 'plug']
        self.agent._port_id_to_vhost = mock.Mock()
        self.agent._plug_tap = mock.Mock()
        self.agent.mirror.get_status = mock.Mock()

        vf = mock.Mock()
        vf.get_status.return_value = None

        dst_vhost = mock.Mock()
        dst_vhost.vf = mock.Mock()
        dst_vhost.vf.get_status = mock.Mock()
        dst_vhost.vf.get_status.return_value = None
        src_vhost = mock.Mock()
        src_vhost.vf = mock.Mock()
        src_vhost.vf.get_status = mock.Mock()
        src_vhost.vf.get_status.return_value = None
        self.agent._port_id_to_vhost.side_effect = [dst_vhost, src_vhost]

        ret = self.agent._do_tap_plug_unplug(tap_flow_id)
        self.assertEqual(ret, None)
        src_vhost.vf.get_status.assert_called()
        self.agent._plug_tap.assert_called_with(tap_flow_id,
                                                'in',
                                                'out',
                                                dst_vhost,
                                                src_vhost)

    @mock.patch('networking_spp.common.etcd_key.tap_info_key')
    def test_do_tap_plug_unplug_con4(self, m_tap_info_key):
        # conditions to exec the 1st else statement and make it return None
        tap_flow_id = 3
        data = {"mirror_in": 'in',
                "mirror_out": 'out',
                "service_port": 'service',
                "source_port": 'source'}
        value = json.dumps(data)
        self.agent.etcd.get.side_effect = [value, 'unplug']
        self.agent._port_id_to_vhost = mock.Mock()
        self.agent._unplug_tap = mock.Mock()
        self.agent.mirror.get_status = mock.Mock()

        dst_vhost = mock.Mock()
        dst_vhost.vf = mock.Mock()
        dst_vhost.vf.get_status = mock.Mock()
        dst_vhost.vf.get_status.return_value = None
        src_vhost = mock.Mock()
        src_vhost.vf = mock.Mock()
        src_vhost.vf.get_status = mock.Mock()
        src_vhost.vf.get_status.return_value = None
        self.agent._port_id_to_vhost.side_effect = [dst_vhost, src_vhost]

        ret = self.agent._do_tap_plug_unplug(tap_flow_id)
        self.assertEqual(ret, None)
        self.agent._unplug_tap.assert_called_with(tap_flow_id,
                                                  'in',
                                                  'out',
                                                  dst_vhost,
                                                  src_vhost)

    @mock.patch('networking_spp.common.etcd_key.tap_info_host_prefix')
    def test_unplug_tap_port_con1(self, m_tap_info_host_prefix):
        # conditions to exec the 1st if statement then continue
        port_id = 4
        data = {"mirror_in": 'in',
                "mirror_out": 'out',
                "service_port": 3,
                "source_port": 5}
        value = json.dumps(data)
        m_tap_info_host_prefix.return_value = 'key'
        self.agent.etcd.get.return_value = None
        self.agent._port_id_to_vhost = mock.Mock()
        self.agent._unplug_tap = mock.Mock()
        self.agent.etcd.get_prefix = mock.Mock()
        self.agent.etcd.get_prefix.return_value = [('key8', value)]
        self.agent._unplug_tap_port(port_id)
        self.agent._port_id_to_vhost.assert_not_called()
        self.agent._unplug_tap.assert_not_called()

    @mock.patch('networking_spp.common.etcd_key.tap_info_host_prefix')
    def test_unplug_tap_port_con2(self, m_tap_info_host_prefix):
        # conditions to exec the 2nd if statement then continue
        port_id = 4
        data = {"mirror_in": 'in',
                "mirror_out": 'out',
                "service_port": 4,
                "source_port": 5}
        value = json.dumps(data)
        m_tap_info_host_prefix.return_value = 'key'
        self.agent.etcd.get.return_value = None
        self.agent._port_id_to_vhost = mock.Mock()
        self.agent._unplug_tap = mock.Mock()
        self.agent.etcd.get_prefix = mock.Mock()
        self.agent.etcd.get_prefix.return_value = [('key8', value)]
        self.agent._port_id_to_vhost.side_effect = [None, 'src_vhost']

        self.agent._unplug_tap_port(port_id)
        self.agent._unplug_tap.assert_not_called()

    @mock.patch('networking_spp.common.etcd_key.tap_info_host_prefix')
    def test_unplug_tap_port_con3(self, m_tap_info_host_prefix):
        # conditions to exec the 3nd if statement
        port_id = 4
        data = {"mirror_in": 'in',
                "mirror_out": 'out',
                "service_port": 4,
                "source_port": 5}
        value = json.dumps(data)
        m_tap_info_host_prefix.return_value = 'key'
        self.agent.etcd.get.return_value = None
        self.agent._port_id_to_vhost = mock.Mock()
        self.agent._unplug_tap = mock.Mock()
        self.agent.etcd.get_prefix = mock.Mock()
        self.agent.etcd.get_prefix.return_value = [('key8', value)]

        dst_vhost = mock.Mock()
        dst_vhost.vf = mock.Mock()
        dst_vhost.vf.get_status = mock.Mock()
        dst_vhost.vf.get_status.return_value = None
        src_vhost = mock.Mock()
        src_vhost.vf = mock.Mock()
        src_vhost.vf.get_status = mock.Mock()
        src_vhost.vf.get_status.return_value = None
        self.agent._port_id_to_vhost.side_effect = [dst_vhost, src_vhost]

        self.agent._unplug_tap_port(port_id)
        src_vhost.vf.get_status.assert_called()
        self.agent._unplug_tap.assert_called_with('8',
                                                  u'in',
                                                  u'out',
                                                  dst_vhost,
                                                  src_vhost)

    @mock.patch('networking_spp.agent.spp_agent.SppAgent._do_tap_plug_unplug')
    @mock.patch('networking_spp.common.etcd_key.tap_action_host_prefix')
    def test_tap_plug_watch_con1(self, m_prefix, m_tap):
        # conditions to exec if statement and for statement twice
        m_prefix.return_value = '/aa/bb/'
        self.agent.etcd.watch_prefix.return_value = [('/aa/bb/123', 'plug'),
                                                     ('/aa/bb/321', 'unplug')]
        self.agent.tap_plug_watch()
        self.assertEqual(m_tap.call_count, 2)
        self.assertEqual(m_tap.call_args_list[0][0][0], '123')
        self.assertEqual(m_tap.call_args_list[1][0][0], '321')

    @mock.patch('networking_spp.agent.spp_agent.SppAgent._do_tap_plug_unplug')
    @mock.patch('networking_spp.common.etcd_key.tap_action_host_prefix')
    def test_tap_plug_watch_con2(self, m_prefix, m_tap):
        # conditions to exec except statement
        m_prefix.return_value = '/aa/bb/'
        self.agent.etcd.watch_prefix.return_value = Exception
        self.agent.port_plug_watch_failed = False
        self.agent.tap_plug_watch()
        self.assertEqual(self.agent.port_plug_watch_failed, True)



# Copyright (c) 2017 NTT
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import random
from unittest import mock

from networking_spp.agent import spp_api
from neutron.tests import base


class SppVfApiTestCase(base.BaseTestCase):

    def setUp(self):
        super(SppVfApiTestCase, self).setUp()
        self.sec_id = random.randint(1, 99)
        self.api_ip_addr = "127.0.0.1"
        self.api_port = 7777
        self.spp_api = spp_api.SppVfApi(self.sec_id, self.api_ip_addr,
                                        self.api_port)

    @mock.patch('requests.request')
    def test_send_request_con1(self, m_request):
        method = "METHOD"
        url = "/url"
        data = {"data": "data"}
        path = self.spp_api.host_url + self.spp_api.proc_url + url
        ret = mock.Mock()
        ret.status_code = 204
        m_request.return_value = ret
        self.spp_api.send_request(method, url, data)
        m_request.assert_called_with(method, path, json=data)

    @mock.patch('networking_spp.agent.spp_api.SppVfApi'
                '.send_request')
    def test_get_status(self, m_send_request):
        m_send_request.return_value = 'info1'
        self.spp_api.get_status()
        self.assertEqual(self.spp_api.info, 'info1')
        m_send_request.assert_called_with("GET", "")

    @mock.patch('networking_spp.agent.spp_api.SppVfApi'
                '.send_request')
    def test_make_component(self, m_send_request):
        comp_name = 'cn'
        core_id = 'ci'
        comp_type = 'ct'
        data = {"name": comp_name, "core": core_id, "type": comp_type}
        self.spp_api.make_component(comp_name, core_id, comp_type)
        m_send_request.assert_called_with("POST", "/components", data)

    @mock.patch('networking_spp.agent.spp_api.SppVfApi'
                '.send_request')
    def test_port_add(self, m_send_request):
        port = 'port'
        direction = 'd'
        comp_name = 'cn'
        data = {"action": "attach", "port": port, "dir": direction}
        path = "/components/%s/ports" % comp_name
        self.spp_api.port_add(port, direction, comp_name)
        m_send_request.assert_called_with("PUT", path, data)

    @mock.patch('networking_spp.agent.spp_api.SppVfApi'
                '.send_request')
    def test_port_add_add_vlantag(self, m_send_request):
        port = 'port'
        direction = 'd'
        comp_name = 'cn'
        op = 'add_vlantag'
        vlan_id = 10
        data = {"action": "attach", "port": port, "dir": direction,
                "vlan": {"operation": "add", "id": vlan_id, "pcp": 0}}
        path = "/components/%s/ports" % comp_name
        self.spp_api.port_add(port, direction, comp_name, op, vlan_id)
        m_send_request.assert_called_with("PUT", path, data)

    @mock.patch('networking_spp.agent.spp_api.SppVfApi'
                '.send_request')
    def test_port_add_del_vlantag(self, m_send_request):
        port = 'port'
        direction = 'd'
        comp_name = 'cn'
        op = 'del_vlantag'
        data = {"action": "attach", "port": port, "dir": direction,
                "vlan": {"operation": "del"}}
        path = "/components/%s/ports" % comp_name
        self.spp_api.port_add(port, direction, comp_name, op)
        m_send_request.assert_called_with("PUT", path, data)

    @mock.patch('networking_spp.agent.spp_api.SppVfApi'
                '.send_request')
    def test_port_del(self, m_send_request):
        port = 'port'
        direction = 'd'
        comp_name = 'cn'
        data = {"action": "detach", "port": port, "dir": direction}
        path = "/components/%s/ports" % comp_name
        self.spp_api.port_del(port, direction, comp_name)
        m_send_request.assert_called_with("PUT", path, data)

    def test_port_exist_con1(self):
        self.spp_api.info = {
            "components": [{"name": "con1",
                            "rx_port": [{"port": "ring:1"}],
                            "tx_port": [{"port": "vhost:1"}]}]}
        self.assertTrue(self.spp_api.port_exist("ring:1", "rx", "con1"))

    def test_port_exist_con2(self):
        self.spp_api.info = {
            "components": [{"name": "con1",
                            "rx_port": [{"port": "ring:1"}],
                            "tx_port": [{"port": "vhost:1"}]}]}
        self.assertFalse(self.spp_api.port_exist("ring:1", "tx", "con1"))

    @mock.patch('networking_spp.agent.spp_api.SppVfApi'
                '.send_request')
    def test_set_classifier_table(self, m_send_request):
        mac = 'mac'
        port = 'port'
        data = {"action": "add", "type": "mac", "mac_address": mac,
                "port": port}
        self.spp_api.set_classifier_table(mac, port)
        m_send_request.assert_called_with("PUT", "/classifier_table", data)

    @mock.patch('networking_spp.agent.spp_api.SppVfApi'
                '.send_request')
    def test_clear_classifier_table(self, m_send_request):
        mac = 'mac'
        port = 'port'
        data = {"action": "del", "type": "mac", "mac_address": mac,
                "port": port}
        self.spp_api.clear_classifier_table(mac, port)
        m_send_request.assert_called_with("PUT", "/classifier_table", data)

    @mock.patch('networking_spp.agent.spp_api.SppVfApi'
                '.send_request')
    def test_set_classifier_table_with_vlan(self, m_send_request):
        mac = 'mac'
        port = 'port'
        vlan_id = 10
        data = {"action": "add", "type": "vlan", "mac_address": mac,
                "port": port, "vlan": vlan_id}
        self.spp_api.set_classifier_table_with_vlan(mac, port, vlan_id)
        m_send_request.assert_called_with("PUT", "/classifier_table", data)

    @mock.patch('networking_spp.agent.spp_api.SppVfApi'
                '.send_request')
    def test_clear_classifier_table_with_vlan(self, m_send_request):
        mac = 'mac'
        port = 'port'
        vlan_id = 10
        data = {"action": "del", "type": "vlan", "mac_address": mac,
                "port": port, "vlan": vlan_id}
        self.spp_api.clear_classifier_table_with_vlan(mac, port, vlan_id)
        m_send_request.assert_called_with("PUT", "/classifier_table", data)


class SppMirrorApiTestCase(base.BaseTestCase):

    def setUp(self):
        super(SppMirrorApiTestCase, self).setUp()
        self.sec_id = random.randint(1, 99)
        self.api_ip_addr = "127.0.0.1"
        self.api_port = 7777
        self.spp_api = spp_api.SppMirrorApi(self.sec_id, self.api_ip_addr,
                                            self.api_port)

    @mock.patch('networking_spp.agent.spp_api.SppMirrorApi'
                '.send_request')
    def test_make_component(self, m_send_request):
        comp_name = 'cn'
        core_id = 'ci'
        data = {"name": comp_name, "core": core_id, "type": "mirror"}
        self.spp_api.make_component(comp_name, core_id)
        m_send_request.assert_called_with("POST", "/components", data)



# Copyright (c) 2017 NTT
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock

from networking_spp.common import etcd_key
from networking_spp.service_drivers import taas
from neutron.tests import base
from neutron_lib import exceptions as n_exc


class SppTaasDriverTestCase(base.BaseTestCase):

    def setUp(self):
        super(SppTaasDriverTestCase, self).setUp()
        self.service_plugin = mock.Mock()
        self.driver = taas.SppTaasDriver(self.service_plugin)
        self.driver.etcd = mock.Mock()

    def test_create_tap_service_con1(self):
        context = mock.Mock()
        context.tap_service = {'port_id': '123'}
        context._plugin_context = mock.Mock()
        port = {'binding:host_id': 'host1'}
        self.driver.etcd.get.return_value = {'key': 'value'}
        self.driver.service_plugin._get_port_details = mock.Mock()
        self.driver.service_plugin._get_port_details.return_value = port

        self.driver.create_tap_service_postcommit(context)

        self.driver.service_plugin._get_port_details.assert_called_with(
            context._plugin_context, '123')

    def test_create_tap_service_con2(self):
        context = mock.Mock()
        context.tap_service = {'port_id': '123'}
        context._plugin_context = mock.Mock()
        port = {'binding:host_id': None}
        self.driver.etcd.get.return_value = {'key': 'value'}
        self.driver.service_plugin._get_port_details = mock.Mock()
        self.driver.service_plugin._get_port_details.return_value = port

        self.assertRaises(taas.PortNotBound,
                          self.driver.create_tap_service_postcommit, context)

    def test_create_tap_service_con3(self):
        context = mock.Mock()
        context.tap_service = {'port_id': '123'}
        context._plugin_context = mock.Mock()
        port = {'binding:host_id': 'host1'}
        self.driver.etcd.get.return_value = None
        self.driver.service_plugin._get_port_details = mock.Mock()
        self.driver.service_plugin._get_port_details.return_value = port

        self.assertRaises(taas.PortNotBound,
                          self.driver.create_tap_service_postcommit, context)

    def test_get_mirror_con1(self):
        prefix = etcd_key.mirror_prefix('host1')
        self.driver.etcd.get_prefix.return_value = [
            (prefix + "0", "tf1"), (prefix + "1", "tf2")]

        ret = self.driver._get_mirror('host1', 'tf3')

        self.assertEqual(None, ret)

    @mock.patch('networking_spp.service_drivers.taas.LOG.debug')
    def test_get_mirror_con2(self, mocked_debug):
        prefix = etcd_key.mirror_prefix('host1')
        self.driver.etcd.get_prefix.return_value = [
            (prefix + "0", "tf1"), (prefix + "1", "None")]

        ret = self.driver._get_mirror('host1', 'tf3')

        self.driver.etcd.replace.assert_called_with(prefix + "1", "None",
                                                    'tf3')
        self.assertEqual("1", ret)
        mocked_debug.assert_called()

    def test_free_mirror(self):
        self.driver._free_mirror('host1', '2')

        self.driver.etcd.put.assert_called_with(
            etcd_key.mirror_key('host1', '2'), 'None')

    def test_create_tap_flow_con1(self):
        context = mock.Mock()
        context.tap_flow = {'source_port': '123'}
        self.driver.service_plugin._get_port_details = mock.Mock()
        port = {'binding:host_id': None}
        self.driver.service_plugin._get_port_details.return_value = port

        self.assertRaises(taas.PortNotBound,
                          self.driver.create_tap_flow_postcommit, context)

    def test_create_tap_flow_con2(self):
        context = mock.Mock()
        context.tap_flow = {'source_port': '123', 'tap_service_id': 'abc'}
        self.driver.service_plugin._get_port_details = mock.Mock()
        port = {'binding:host_id': 'host1'}
        self.driver.service_plugin._get_port_details.return_value = port
        self.driver.service_plugin.get_tap_service = mock.Mock()
        ts = {'port_id': '456'}
        self.driver.service_plugin.get_tap_service.return_value = ts
        self.driver.etcd.get.side_effect = ['aaa', None]

        self.assertRaises(taas.NotSameHost,
                          self.driver.create_tap_flow_postcommit, context)

    def test_create_tap_flow_con3(self):
        context = mock.Mock()
        context.tap_flow = {'source_port': '123', 'tap_service_id': 'abc',
                            'id': 'tf1', 'direction': 'IN'}
        self.driver.service_plugin._get_port_details = mock.Mock()
        port = {'binding:host_id': 'host1'}
        self.driver.service_plugin._get_port_details.return_value = port
        self.driver.service_plugin.get_tap_service = mock.Mock()
        ts = {'port_id': '456'}
        self.driver.service_plugin.get_tap_service.return_value = ts
        self.driver.etcd.get.side_effect = ['aaa', 'aaa']
        self.driver._get_mirror = mock.Mock(return_value=None)

        self.assertRaises(taas.NoMirrorAvailable,
                          self.driver.create_tap_flow_postcommit, context)

    def test_create_tap_flow_con4(self):
        context = mock.Mock()
        context.tap_flow = {'source_port': '123', 'tap_service_id': 'abc',
                            'id': 'tf1', 'direction': 'OUT'}
        self.driver.service_plugin._get_port_details = mock.Mock()
        port = {'binding:host_id': 'host1'}
        self.driver.service_plugin._get_port_details.return_value = port
        self.driver.service_plugin.get_tap_service = mock.Mock()
        ts = {'port_id': '456'}
        self.driver.service_plugin.get_tap_service.return_value = ts
        self.driver.etcd.get.side_effect = ['aaa', 'aaa']
        self.driver._get_mirror = mock.Mock(return_value=None)

        self.assertRaises(taas.NoMirrorAvailable,
                          self.driver.create_tap_flow_postcommit, context)

    def test_create_tap_flow_con5(self):
        context = mock.Mock()
        context.tap_flow = {'source_port': '123', 'tap_service_id': 'abc',
                            'id': 'tf1', 'direction': 'BOTH'}
        self.driver.service_plugin._get_port_details = mock.Mock()
        port = {'binding:host_id': 'host1'}
        self.driver.service_plugin._get_port_details.return_value = port
        self.driver.service_plugin.get_tap_service = mock.Mock()
        ts = {'port_id': '456'}
        self.driver.service_plugin.get_tap_service.return_value = ts
        self.driver.etcd.get.side_effect = ['aaa', 'aaa']
        self.driver._get_mirror = mock.Mock()
        self.driver._get_mirror.side_effect = ['0', None]
        self.driver._free_mirror = mock.Mock()

        self.assertRaises(taas.NoMirrorAvailable,
                          self.driver.create_tap_flow_postcommit, context)
        self.driver._free_mirror.assert_called_once()

    @mock.patch('networking_spp.service_drivers.taas.LOG.debug')
    def test_create_tap_flow_con6(self, mocked_debug):
        context = mock.Mock()
        context.tap_flow = {'source_port': '123', 'tap_service_id': 'abc',
                            'id': 'tf1', 'direction': 'BOTH'}
        self.driver.service_plugin._get_port_details = mock.Mock()
        port = {'binding:host_id': 'host1'}
        self.driver.service_plugin._get_port_details.return_value = port
        self.driver.service_plugin.get_tap_service = mock.Mock()
        ts = {'port_id': '456'}
        self.driver.service_plugin.get_tap_service.return_value = ts
        self.driver.etcd.get.side_effect = ['aaa', 'aaa']
        self.driver._get_mirror = mock.Mock()
        self.driver._get_mirror.side_effect = ['0', '1']
        value = {"service_port": "456", "source_port": "123",
                 "mirror_in": 0, "mirror_out": 1}
        self.driver.etcd.watch_once = mock.Mock(return_value=('k', 'up'))

        self.driver.create_tap_flow_postcommit(context)

        k, v = self.driver.etcd.put.call_args_list[0][0]
        self.assertEqual(etcd_key.tap_info_key('host1', 'tf1'), k)
        self.assertEqual(value, json.loads(v))
        k, v = self.driver.etcd.put.call_args_list[1][0]
        self.assertEqual(etcd_key.tap_action_key('host1', 'tf1'), k)
        self.assertEqual("plug", v)
        mocked_debug.assert_called()

    @mock.patch('networking_spp.service_drivers.taas.LOG.warning')
    def test_create_tap_flow_con7(self, mocked_warning):
        context = mock.Mock()
        context.tap_flow = {'source_port': '123', 'tap_service_id': 'abc',
                            'id': 'tf1', 'direction': 'OUT'}
        self.driver.service_plugin._get_port_details = mock.Mock()
        port = {'binding:host_id': 'host1'}
        self.driver.service_plugin._get_port_details.return_value = port
        self.driver.service_plugin.get_tap_service = mock.Mock()
        ts = {'port_id': '456'}
        self.driver.service_plugin.get_tap_service.return_value = ts
        self.driver.etcd.get.side_effect = ['aaa', 'aaa']
        self.driver._get_mirror = mock.Mock(return_value='2')
        value = {"service_port": "456", "source_port": "123",
                 "mirror_in": None, "mirror_out": 2}
        self.driver.etcd.watch_once = mock.Mock(return_value=(None, None))

        self.driver.create_tap_flow_postcommit(context)

        k, v = self.driver.etcd.put.call_args_list[0][0]
        self.assertEqual(etcd_key.tap_info_key('host1', 'tf1'), k)
        self.assertEqual(value, json.loads(v))
        k, v = self.driver.etcd.put.call_args_list[1][0]
        self.assertEqual(etcd_key.tap_action_key('host1', 'tf1'), k)
        self.assertEqual("plug", v)
        mocked_warning.assert_called()

    def test_delete_tap_flow_con1(self):
        context = mock.Mock()
        context.tap_flow = {'source_port': '123', 'id': 'tf1'}
        self.driver.service_plugin._get_port_details = mock.Mock()
        port = {'binding:host_id': None}
        self.driver.service_plugin._get_port_details.return_value = port

        self.driver.delete_tap_flow_postcommit(context)

        self.driver.etcd.get.assert_not_called()

    def test_delete_tap_flow_con2(self):
        context = mock.Mock()
        context.tap_flow = {'source_port': '123', 'id': 'tf1'}
        self.driver.service_plugin._get_port_details = mock.Mock(
            side_effect=n_exc.PortNotFound(port_id="port_id"))

        self.driver.delete_tap_flow_postcommit(context)

        self.driver.etcd.get.assert_not_called()

    def test_delete_tap_flow_con3(self):
        context = mock.Mock()
        context.tap_flow = {'source_port': '123', 'id': 'tf1'}
        self.driver.service_plugin._get_port_details = mock.Mock()
        port = {'binding:host_id': 'host1'}
        self.driver.service_plugin._get_port_details.return_value = port
        self.driver.etcd.get.return_value = 'aaa'

        self.driver.delete_tap_flow_postcommit(context)

        key = etcd_key.tap_action_key('host1', 'tf1')
        self.driver.etcd.put.assert_called_with(key, "unplug")
